# Copyright 2015 Cloudbase Solutions SRL
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Unit tests for the Hyper-V Mechanism Driver.
"""

import mock

from networking_hyperv.neutron import constants
from networking_hyperv.neutron.ml2 import mech_hyperv
from networking_hyperv.tests import base


class TestHypervMechanismDriver(base.BaseTestCase):

    def setUp(self):
        super(TestHypervMechanismDriver, self).setUp()
        self.mech_hyperv = mech_hyperv.HypervMechanismDriver()

    def test_get_allowed_network_types(self):
        agent = {'configurations': {'tunnel_types': []}}
        actual_net_types = self.mech_hyperv.get_allowed_network_types(agent)

        network_types = [constants.TYPE_LOCAL, constants.TYPE_FLAT,
                         constants.TYPE_VLAN]
        self.assertEqual(network_types, actual_net_types)

    def test_get_allowed_network_types_nvgre(self):
        agent = {'configurations': {'tunnel_types': [constants.TYPE_NVGRE]}}
        actual_net_types = self.mech_hyperv.get_allowed_network_types(agent)

        network_types = [constants.TYPE_LOCAL, constants.TYPE_FLAT,
                         constants.TYPE_VLAN, constants.TYPE_NVGRE]
        self.assertEqual(network_types, actual_net_types)

    def test_get_mappings(self):
        agent = {'configurations': {
            'vswitch_mappings': [mock.sentinel.mapping]}}
        mappings = self.mech_hyperv.get_mappings(agent)
        self.assertEqual([mock.sentinel.mapping], mappings)

    def test_physnet_in_mappings(self):
        physnet = 'test_physnet'
        match_mapping = '.*'
        different_mapping = 'fake'

        pattern_matched = self.mech_hyperv.physnet_in_mappings(
            physnet, [match_mapping])
        self.assertTrue(pattern_matched)

        pattern_matched = self.mech_hyperv.physnet_in_mappings(
            physnet, [different_mapping])
        self.assertFalse(pattern_matched)

        pattern_matched = self.mech_hyperv.physnet_in_mappings(
            physnet, [different_mapping, match_mapping])
        self.assertTrue(pattern_matched)



# Copyright 2017 Cloudbase Solutions SRL
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Unit tests for the Neutron HNV L2 Agent.
"""

import sys

import mock

from networking_hyperv.neutron.agent import hnv_neutron_agent as hnv_agent
from networking_hyperv.neutron import constants
from networking_hyperv.tests import base as test_base


class TestHNVAgent(test_base.HyperVBaseTestCase):

    _autospec_classes = [
        hnv_agent.neutron_client.NeutronAPIClient,
    ]

    @mock.patch.object(hnv_agent.HNVAgent, "_setup")
    @mock.patch.object(hnv_agent.HNVAgent, "_setup_rpc")
    @mock.patch.object(hnv_agent.HNVAgent, "_set_agent_state")
    def _get_agent(self, mock_set_agent_state, mock_setup_rpc, mock_setup):
        return hnv_agent.HNVAgent()

    def setUp(self):
        super(TestHNVAgent, self).setUp()

        self.agent = self._get_agent()
        self.agent._utils = mock.Mock(autospec=self.agent._utils)

    def test_get_agent_configurations(self):
        self.config(logical_network=mock.sentinel.logical_network,
                    group="HNV")
        self.agent._physical_network_mappings = mock.sentinel.mappings

        agent_configurations = self.agent._get_agent_configurations()

        expected_keys = ["logical_network", "vswitch_mappings",
                         "devices", "l2_population", "tunnel_types",
                         "bridge_mappings", "enable_distributed_routing"]
        self.assertEqual(sorted(expected_keys),
                         sorted(agent_configurations.keys()))
        self.assertEqual(mock.sentinel.mappings,
                         agent_configurations["vswitch_mappings"])
        self.assertEqual(str(mock.sentinel.logical_network),
                         agent_configurations["logical_network"])

    @mock.patch.object(hnv_agent.HNVAgent, "_get_vswitch_name")
    def test_provision_network(self, mock_get_vswitch_name):
        self.agent._provision_network(mock.sentinel.port_id,
                                      mock.sentinel.net_uuid,
                                      mock.sentinel.network_type,
                                      mock.sentinel.physical_network,
                                      mock.sentinel.segmentation_id)

        mock_get_vswitch_name.assert_called_once_with(
            mock.sentinel.network_type,
            mock.sentinel.physical_network)

        vswitch_map = self.agent._network_vswitch_map[mock.sentinel.net_uuid]
        self.assertEqual(mock.sentinel.network_type,
                         vswitch_map['network_type'])
        self.assertEqual(mock_get_vswitch_name.return_value,
                         vswitch_map['vswitch_name'])
        self.assertEqual(mock.sentinel.segmentation_id,
                         vswitch_map['vlan_id'])

    @mock.patch.object(hnv_agent.hyperv_base.Layer2Agent, '_port_bound')
    def test_port_bound(self, mock_super_port_bound):
        self.agent._port_bound(
            mock.sentinel.port_id, mock.sentinel.network_id,
            mock.sentinel.network_type, mock.sentinel.physical_network,
            mock.sentinel.segmentation_id, mock.sentinel.port_security_enabled,
            mock.sentinel.set_port_sriov)

        mock_super_port_bound.assert_called_once_with(
            mock.sentinel.port_id, mock.sentinel.network_id,
            mock.sentinel.network_type, mock.sentinel.physical_network,
            mock.sentinel.segmentation_id, mock.sentinel.port_security_enabled,
            mock.sentinel.set_port_sriov)
        mock_neutron_client = self.agent._neutron_client
        mock_neutron_client.get_port_profile_id.assert_called_once_with(
            mock.sentinel.port_id)
        self.agent._utils.set_vswitch_port_profile_id.assert_called_once_with(
            switch_port_name=mock.sentinel.port_id,
            profile_id=mock_neutron_client.get_port_profile_id.return_value,
            profile_data=constants.PROFILE_DATA,
            profile_name=constants.PROFILE_NAME,
            net_cfg_instance_id=constants.NET_CFG_INSTANCE_ID,
            cdn_label_id=constants.CDN_LABEL_ID,
            cdn_label_string=constants.CDN_LABEL_STRING,
            vendor_id=constants.VENDOR_ID,
            vendor_name=constants.VENDOR_NAME)


class TestMain(test_base.BaseTestCase):

    @mock.patch.object(hnv_agent, 'HNVAgent')
    @mock.patch.object(hnv_agent, 'common_config')
    @mock.patch.object(hnv_agent, 'neutron_config')
    def test_main(self, mock_config, mock_common_config, mock_hnv_agent):
        hnv_agent.main()

        mock_config.register_agent_state_opts_helper.assert_called_once_with(
            hnv_agent.CONF)
        mock_common_config.init.assert_called_once_with(sys.argv[1:])
        mock_config.setup_logging.assert_called_once_with()
        mock_hnv_agent.assert_called_once_with()
        mock_hnv_agent.return_value.daemon_loop.assert_called_once_with()



# Copyright 2017 Cloudbase Solutions Srl
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Unit tests for Neutron base agent.
"""

import mock
from neutron.agent import rpc as agent_rpc

from networking_hyperv.neutron.agent import base as agent_base
from networking_hyperv.tests import base as test_base


class _BaseAgent(agent_base.BaseAgent):

    def _get_agent_configurations(self):
        pass

    def _setup_rpc(self):
        pass

    def _work(self):
        pass


class TestBaseAgent(test_base.HyperVBaseTestCase):

    def setUp(self):
        super(TestBaseAgent, self).setUp()

        self._agent = _BaseAgent()

        self._agent._agent_id = mock.sentinel.agent_id
        self._agent._context = mock.sentinel.admin_context
        self._agent._state_rpc = mock.MagicMock(
            autospec=agent_rpc.PluginReportStateAPI)

    def test_set_agent_state(self):
        self._agent._agent_state = {}
        self._agent._host = mock.sentinel.host

        self._agent._set_agent_state()

        expected_keys = ["binary", "host", "configurations", "agent_type",
                         "topic", "start_flag"]
        self.assertEqual(sorted(expected_keys),
                         sorted(self._agent._agent_state.keys()))
        self.assertEqual(mock.sentinel.host, self._agent._agent_state["host"])

    @mock.patch('time.time')
    @mock.patch('time.sleep')
    @mock.patch.object(_BaseAgent, '_work')
    @mock.patch.object(_BaseAgent, '_prologue')
    def test_daemon_loop(self, mock_prologue, mock_work,
                         mock_sleep, mock_time):
        mock_work.side_effect = [Exception()]
        mock_time.side_effect = [1, 3, KeyboardInterrupt]

        self.assertRaises(KeyboardInterrupt, self._agent.daemon_loop)

        mock_prologue.assert_called_once_with()

    def test_report_state(self):
        self._agent._agent_state = {'start_flag': True}

        self._agent._report_state()

        self.assertNotIn('start_flag', self._agent._agent_state)

    def test_report_state_exception(self):
        self._agent._agent_state = {'start_flag': True}
        self._agent._state_rpc.report_state.side_effect = Exception

        self._agent._report_state()

        self._agent._state_rpc.report_state.assert_called_once_with(
            self._agent._context, {'start_flag': True})
        self.assertTrue(self._agent._agent_state['start_flag'])



# Copyright 2017 Cloudbase Solutions SRL
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import hashlib
import hmac
import sys

import httplib2
from neutron.agent import rpc as agent_rpc
from neutron.common import config as common_config
from neutron.conf.agent import common as neutron_config
from neutron.conf.agent.metadata import config as meta_config
from neutron import wsgi
from neutron_lib.agent import topics
from neutron_lib import constants
from neutron_lib import context
from oslo_log import log as logging
from oslo_service import loopingcall
from oslo_utils import encodeutils
from oslo_utils import uuidutils
import six
import six.moves.urllib.parse as urlparse
import webob

from networking_hyperv.common.i18n import _, _LW, _LE   # noqa
from networking_hyperv.neutron.agent import base as base_agent
from networking_hyperv.neutron import config
from networking_hyperv.neutron import neutron_client

CONF = config.CONF
LOG = logging.getLogger(__name__)


class _MetadataProxyHandler(object):

    def __init__(self):
        self._context = context.get_admin_context_without_session()
        self._neutron_client = neutron_client.NeutronAPIClient()

    @webob.dec.wsgify(RequestClass=webob.Request)
    def __call__(self, req):
        try:
            return self._proxy_request(req)
        except Exception:
            LOG.exception("Unexpected error.")
            msg = _('An unknown error has occurred. '
                    'Please try your request again.')
            explanation = six.text_type(msg)
            return webob.exc.HTTPInternalServerError(explanation=explanation)

    def _get_port_profile_id(self, request):
        """Get the port profile ID from the request path."""

        # Note(alexcoman): The port profile ID can be found as suffix
        # in request path.
        port_profile_id = request.path.split("/")[-1].strip()
        if uuidutils.is_uuid_like(port_profile_id):
            LOG.debug("The instance id was found in request path.")
            return port_profile_id

        LOG.debug("Failed to get the instance id from the request.")
        return None

    def _get_instance_id(self, port_profile_id):
        tenant_id = None
        instance_id = None
        ports = self._neutron_client.get_network_ports()
        for port in ports:
            vif_details = port.get("binding:vif_details", {})
            profile_id = vif_details.get("port_profile_id")
            if profile_id and profile_id == port_profile_id:
                tenant_id = port["tenant_id"]
                # Note(alexcoman): The port["device_id"] is actually the
                # Nova instance_id.
                instance_id = port["device_id"]
                break
        else:
            LOG.debug("Failed to get the port information.")

        return tenant_id, instance_id

    def _sign_instance_id(self, instance_id):
        secret = CONF.metadata_proxy_shared_secret
        secret = encodeutils.to_utf8(secret)
        instance_id = encodeutils.to_utf8(instance_id)
        return hmac.new(secret, instance_id, hashlib.sha256).hexdigest()

    def _get_headers(self, port_profile_id):
        tenant_id, instance_id = self._get_instance_id(port_profile_id)
        if not (tenant_id and instance_id):
            return None

        headers = {
            'X-Instance-ID': instance_id,
            'X-Tenant-ID': tenant_id,
            'X-Instance-ID-Signature': self._sign_instance_id(instance_id),
        }
        return headers

    def _proxy_request(self, request):
        LOG.debug("Request: %s", request)
        port_profile_id = self._get_port_profile_id(request)
        if not port_profile_id:
            return webob.exc.HTTPNotFound()

        headers = self._get_headers(port_profile_id)
        if not headers:
            return webob.exc.HTTPNotFound()

        LOG.debug("Trying to proxy the request.")
        nova_url = '%s:%s' % (CONF.nova_metadata_host,
                              CONF.nova_metadata_port)
        allow_insecure = CONF.nova_metadata_insecure

        http_request = httplib2.Http(
            ca_certs=CONF.auth_ca_cert,
            disable_ssl_certificate_validation=allow_insecure
        )
        if CONF.nova_client_cert and CONF.nova_client_priv_key:
            http_request.add_certificate(
                key=CONF.nova_client_priv_key,
                cert=CONF.nova_client_cert,
                domain=nova_url)

        url = urlparse.urlunsplit((
            CONF.nova_metadata_protocol, nova_url,
            request.path_info, request.query_string, ''))

        response, content = http_request.request(
            url.replace(port_profile_id, ""),
            method=request.method, headers=headers,
            body=request.body)

        LOG.debug("Response [%s]: %s", response.status, content)
        if response.status == 200:
            request.response.content_type = response['content-type']
            request.response.body = content
            return request.response
        elif response.status == 403:
            LOG.warning('The remote metadata server responded with Forbidden. '
                        'This response usually occurs when shared secrets do '
                        'not match.')
            return webob.exc.HTTPForbidden()
        elif response.status == 400:
            return webob.exc.HTTPBadRequest()
        elif response.status == 404:
            return webob.exc.HTTPNotFound()
        elif response.status == 409:
            return webob.exc.HTTPConflict()
        elif response.status == 500:
            message = _(
                "Remote metadata server experienced an internal server error."
            )
            LOG.warning(message)
            return webob.exc.HTTPInternalServerError(explanation=message)
        else:
            message = _("The HNV Metadata proxy experienced an internal"
                        " server error.")
            LOG.warning('Unexpected response code: %s', response.status)
            return webob.exc.HTTPInternalServerError(explanation=message)


class MetadataProxy(base_agent.BaseAgent):

    _AGENT_BINARY = 'neutron-hnv-metadata-proxy'
    _AGENT_TYPE = constants.AGENT_TYPE_METADATA
    _AGENT_TOPIC = 'N/A'

    def __init__(self):
        super(MetadataProxy, self).__init__()
        self._set_agent_state()
        self._setup_rpc()

    def _setup_rpc(self):
        """Setup the RPC client for the current agent."""
        self._state_rpc = agent_rpc.PluginReportStateAPI(topics.REPORTS)
        report_interval = CONF.AGENT.report_interval
        if report_interval:
            heartbeat = loopingcall.FixedIntervalLoopingCall(
                self._report_state)
            heartbeat.start(interval=report_interval)

    def _get_agent_configurations(self):
        return {
            'nova_metadata_host': CONF.nova_metadata_host,
            'nova_metadata_port': CONF.nova_metadata_port,
            'log_agent_heartbeats': CONF.AGENT.log_agent_heartbeats,
        }

    def _work(self):
        """Start the neutron-hnv-metadata-proxy agent."""
        server = wsgi.Server(
            name=self._AGENT_BINARY,
            num_threads=CONF.AGENT.worker_count)
        server.start(
            application=_MetadataProxyHandler(),
            port=CONF.bind_port,
            host=CONF.bind_host)
        server.wait()

    def run(self):
        self._prologue()
        try:
            self._work()
        except Exception:
            LOG.exception("Error in agent.")


def register_config_opts():
    neutron_config.register_agent_state_opts_helper(CONF)
    meta_config.register_meta_conf_opts(
        meta_config.METADATA_PROXY_HANDLER_OPTS)


def main():
    """The entry point for neutron-hnv-metadata-proxy."""
    register_config_opts()
    common_config.init(sys.argv[1:])
    neutron_config.setup_logging()
    proxy = MetadataProxy()
    proxy.run()



# Copyright 2015 Cloudbase Solutions SRL
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from os_win import utilsfactory
from oslo_log import log as logging
import six
import uuid

from networking_hyperv.common.i18n import _LI, _LW, _LE  # noqa
from networking_hyperv.neutron import config
from networking_hyperv.neutron import constants
from networking_hyperv.neutron import hyperv_agent_notifier
from networking_hyperv.neutron import neutron_client

CONF = config.CONF
LOG = logging.getLogger(__name__)


class HyperVNvgreOps(object):

    def __init__(self, physical_networks):
        self.topic = constants.AGENT_TOPIC
        self._vswitch_ips = {}
        self._tunneling_agents = {}
        self._nvgre_ports = []
        self._network_vsids = {}

        self._hyperv_utils = utilsfactory.get_networkutils()
        self._nvgre_utils = utilsfactory.get_nvgreutils()
        self._n_client = neutron_client.NeutronAPIClient()

        self._init_nvgre(physical_networks)

    def init_notifier(self, context, rpc_client):
        self.context = context
        self._notifier = hyperv_agent_notifier.AgentNotifierApi(
            self.topic, rpc_client)

    def _init_nvgre(self, physical_networks):
        for network in physical_networks:
            LOG.info("Adding provider route and address for network: %s",
                     network)
            self._nvgre_utils.create_provider_route(network)
            self._nvgre_utils.create_provider_address(
                network, CONF.NVGRE.provider_vlan_id)
            ip_addr, length = self._nvgre_utils.get_network_iface_ip(network)
            self._vswitch_ips[network] = ip_addr

    def _refresh_tunneling_agents(self):
        self._tunneling_agents.update(self._n_client.get_tunneling_agents())

    def lookup_update(self, kwargs):
        lookup_ip = kwargs.get('lookup_ip')
        lookup_details = kwargs.get('lookup_details')

        LOG.info("Lookup Received: %(lookup_ip)s, %(lookup_details)s",
                 {'lookup_ip': lookup_ip, 'lookup_details': lookup_details})
        if not lookup_ip or not lookup_details:
            return

        self._register_lookup_record(lookup_ip,
                                     lookup_details['customer_addr'],
                                     lookup_details['mac_addr'],
                                     lookup_details['customer_vsid'])

    def tunnel_update(self, context, tunnel_ip, tunnel_type):
        if tunnel_type != constants.TYPE_NVGRE:
            return
        self._notifier.tunnel_update(context, CONF.NVGRE.provider_tunnel_ip,
                                     tunnel_type)

    def _register_lookup_record(self, prov_addr, cust_addr, mac_addr, vsid):
        LOG.info('Creating LookupRecord: VSID: %(vsid)s MAC: %(mac_addr)s '
                 'Customer IP: %(cust_addr)s Provider IP: %(prov_addr)s',
                 dict(vsid=vsid,
                      mac_addr=mac_addr,
                      cust_addr=cust_addr,
                      prov_addr=prov_addr))

        self._nvgre_utils.create_lookup_record(
            prov_addr, cust_addr, mac_addr, vsid)

    def bind_nvgre_port(self, segmentation_id, network_name, port_id):
        mac_addr = self._hyperv_utils.get_vnic_mac_address(port_id)
        provider_addr = self._nvgre_utils.get_network_iface_ip(network_name)[0]
        customer_addr = self._n_client.get_port_ip_address(port_id)

        if not provider_addr or not customer_addr:
            LOG.warning('Cannot bind NVGRE port. Could not determine '
                        'provider address (%(prov_addr)s) or customer '
                        'address (%(cust_addr)s).',
                        {'prov_addr': provider_addr,
                         'cust_addr': customer_addr})
            return

        LOG.info('Binding VirtualSubnetID %(segmentation_id)s '
                 'to switch port %(port_id)s',
                 dict(segmentation_id=segmentation_id, port_id=port_id))
        self._hyperv_utils.set_vswitch_port_vsid(segmentation_id, port_id)

        # normal lookup record.
        self._register_lookup_record(
            provider_addr, customer_addr, mac_addr, segmentation_id)

        # lookup record for dhcp requests.
        self._register_lookup_record(
            self._vswitch_ips[network_name], constants.IPV4_DEFAULT,
            mac_addr, segmentation_id)

        LOG.info('Fanning out LookupRecord...')
        self._notifier.lookup_update(self.context,
                                     provider_addr,
                                     {'customer_addr': customer_addr,
                                      'mac_addr': mac_addr,
                                      'customer_vsid': segmentation_id})

    def bind_nvgre_network(self, segmentation_id, net_uuid, vswitch_name):
        subnets = self._n_client.get_network_subnets(net_uuid)
        if len(subnets) > 1:
            LOG.warning("Multiple subnets in the same network is not "
                        "supported.")
        subnet = subnets[0]
        try:
            cidr, gw = self._n_client.get_network_subnet_cidr_and_gateway(
                subnet)

            cust_route_string = vswitch_name + cidr + str(segmentation_id)
            rdid_uuid = str(uuid.uuid5(uuid.NAMESPACE_X500, cust_route_string))
            self._create_customer_routes(segmentation_id, cidr, gw, rdid_uuid)

        except Exception as ex:
            LOG.error("Exception caught: %s", ex)

        self._network_vsids[net_uuid] = segmentation_id
        self.refresh_nvgre_records(network_id=net_uuid)
        self._notifier.tunnel_update(
            self.context, CONF.NVGRE.provider_tunnel_ip, segmentation_id)

    def _create_customer_routes(self, segmentation_id, cidr, gw, rdid_uuid):
        self._nvgre_utils.clear_customer_routes(segmentation_id)

        # create cidr -> 0.0.0.0/0 customer route
        self._nvgre_utils.create_customer_route(
            segmentation_id, cidr, constants.IPV4_DEFAULT, rdid_uuid)

        if not gw:
            LOG.info('Subnet does not have gateway configured. Skipping.')
        elif gw.split('.')[-1] == '1':
            LOG.error('Subnet has unsupported gateway IP ending in 1: '
                      '%s. Any other gateway IP is supported.', gw)
        else:
            # create 0.0.0.0/0 -> gateway customer route
            self._nvgre_utils.create_customer_route(
                segmentation_id, '%s/0' % constants.IPV4_DEFAULT, gw,
                rdid_uuid)

            # create metadata address -> gateway customer route
            metadata_addr = '%s/32' % CONF.AGENT.neutron_metadata_address
            self._nvgre_utils.create_customer_route(
                segmentation_id, metadata_addr, gw, rdid_uuid)

    def refresh_nvgre_records(self, **kwargs):
        self._refresh_tunneling_agents()
        ports = self._n_client.get_network_ports(**kwargs)

        # process ports that were not processed yet.
        # process ports that are bound to tunneling_agents.
        ports = [p for p in ports if p['id'] not in self._nvgre_ports and
                 p['binding:host_id'] in self._tunneling_agents and
                 p['network_id'] in six.iterkeys(self._network_vsids)]

        for port in ports:
            tunneling_ip = self._tunneling_agents[port['binding:host_id']]
            customer_addr = port['fixed_ips'][0]['ip_address']
            mac_addr = port['mac_address'].replace(':', '')
            segmentation_id = self._network_vsids[port['network_id']]
            try:
                self._register_lookup_record(
                    tunneling_ip, customer_addr, mac_addr, segmentation_id)

                self._nvgre_ports.append(port['id'])
            except Exception as ex:
                LOG.error("Exception while adding lookup_record: %(ex)s. "
                          "VSID: %(vsid)s MAC: %(mac_address)s Customer "
                          "IP:%(cust_addr)s Provider IP: %(prov_addr)s",
                          dict(ex=ex,
                               vsid=segmentation_id,
                               mac_address=mac_addr,
                               cust_addr=customer_addr,
                               prov_addr=tunneling_ip))



# Copyright 2017 Cloudbase Solutions Srl
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron.api.rpc.callbacks import events
from neutron.api.rpc.handlers import resources_rpc
from neutron.services.trunk.rpc import agent as trunk_rpc
from neutron_lib.services.trunk import constants as t_const
from os_win import constants as os_win_const
from os_win import utilsfactory
from oslo_log import log as logging
import oslo_messaging

from networking_hyperv.common.i18n import _LI, _LE  # noqa

LOG = logging.getLogger(__name__)


class HyperVTrunkDriver(trunk_rpc.TrunkSkeleton):
    """Driver responsible for handling trunk/subport/port events.

    Receives data model events from the neutron server and uses them to setup
    VLAN trunks for Hyper-V vSwitch ports.
    """

    def __init__(self, context):
        super(HyperVTrunkDriver, self).__init__()
        self._context = context
        self._utils = utilsfactory.get_networkutils()
        self._trunk_rpc = trunk_rpc.TrunkStub()

        # Map between trunk.id and trunk.
        self._trunks = {}

    def handle_trunks(self, trunks, event_type):
        """Trunk data model change from the server."""

        LOG.debug("Trunks event received: %(event_type)s. Trunks: %(trunks)s",
                  {'event_type': event_type, 'trunks': trunks})

        if event_type == events.DELETED:
            # The port trunks have been deleted. Remove them from cache.
            for trunk in trunks:
                self._trunks.pop(trunk.id, None)
        else:
            for trunk in trunks:
                self._trunks[trunk.id] = trunk
                self._setup_trunk(trunk)

    def handle_subports(self, subports, event_type):
        """Subport data model change from the server."""

        LOG.debug("Subports event received: %(event_type)s. "
                  "Subports: %(subports)s",
                  {'event_type': event_type, 'subports': subports})

        # update the cache.
        if event_type == events.CREATED:
            for subport in subports:
                trunk = self._trunks.get(subport['trunk_id'])
                if trunk:
                    trunk.sub_ports.append(subport)
        elif event_type == events.DELETED:
            for subport in subports:
                trunk = self._trunks.get(subport['trunk_id'])
                if trunk and subport in trunk.sub_ports:
                    trunk.sub_ports.remove(subport)

        # update the bound trunks.
        affected_trunk_ids = set([s['trunk_id'] for s in subports])
        for trunk_id in affected_trunk_ids:
            trunk = self._trunks.get(trunk_id)
            if trunk:
                self._setup_trunk(trunk)

    def bind_vlan_port(self, port_id, segmentation_id):
        trunk = self._fetch_trunk(port_id)
        if not trunk:
            # No trunk found. No VLAN IDs to set in trunk mode.
            self._set_port_vlan(port_id, segmentation_id)
            return

        self._setup_trunk(trunk, segmentation_id)

    def _fetch_trunk(self, port_id, context=None):
        context = context or self._context
        try:
            trunk = self._trunk_rpc.get_trunk_details(context, port_id)
            LOG.debug("Found trunk for port_id %(port_id)s: %(trunk)s",
                      {'port_id': port_id, 'trunk': trunk})

            # cache it.
            self._trunks[trunk.id] = trunk
            return trunk
        except resources_rpc.ResourceNotFound:
            return None
        except oslo_messaging.RemoteError as ex:
            if 'CallbackNotFound' not in str(ex):
                raise
            LOG.debug("Trunk plugin disabled on server. Assuming port %s is "
                      "not a trunk.", port_id)
            return None

    def _setup_trunk(self, trunk, vlan_id=None):
        """Sets up VLAN trunk and updates the trunk status."""

        LOG.info('Binding trunk port: %s.', trunk)
        try:
            # bind sub_ports to host.
            self._trunk_rpc.update_subport_bindings(self._context,
                                                    trunk.sub_ports)

            vlan_trunk = [s.segmentation_id for s in trunk.sub_ports]
            self._set_port_vlan(trunk.port_id, vlan_id, vlan_trunk)

            self._trunk_rpc.update_trunk_status(self._context, trunk.id,
                                                t_const.TRUNK_ACTIVE_STATUS)
        except Exception:
            # something broke
            LOG.exception("Failure setting up subports for %s", trunk.port_id)
            self._trunk_rpc.update_trunk_status(self._context, trunk.id,
                                                t_const.TRUNK_DEGRADED_STATUS)

    def _set_port_vlan(self, port_id, vlan_id, vlan_trunk=None):
        LOG.info('Binding VLAN ID: %(vlan_id)s, VLAN trunk: '
                 '%(vlan_trunk)s to switch port %(port_id)s',
                 dict(vlan_id=vlan_id, vlan_trunk=vlan_trunk, port_id=port_id))

        op_mode = (os_win_const.VLAN_MODE_TRUNK if vlan_trunk else
                   os_win_const.VLAN_MODE_ACCESS)
        self._utils.set_vswitch_port_vlan_id(
            vlan_id,
            port_id,
            operation_mode=op_mode,
            trunk_vlans=vlan_trunk)



# Copyright 2017 Cloudbase Solutions Srl
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Unit tests for Windows Hyper-V QoS Driver.
"""

import mock
from neutron_lib.services.qos import constants as qos_consts

from networking_hyperv.neutron.qos import qos_driver
from networking_hyperv.tests import base


class TestQosHyperVAgentDriver(base.BaseTestCase):
    @mock.patch.object(qos_driver.QosHyperVAgentDriver, '__init__',
                       lambda *args, **kwargs: None)
    def setUp(self):
        super(TestQosHyperVAgentDriver, self).setUp()
        self.driver = qos_driver.QosHyperVAgentDriver()
        self.driver.initialize()
        self.driver._utils = mock.Mock(autospec=self.driver._utils)

    @mock.patch.object(qos_driver, 'networkutils')
    def test_initialize(self, mock_networkutils):
        self.driver.initialize()
        mock_networkutils.NetworkUtils.assert_called_once_with()

    @mock.patch.object(qos_driver.QosHyperVAgentDriver, '_get_policy_values')
    def test_create(self, mock_get_policy_values):
        self.driver.create({'port_id': mock.sentinel.port_id},
                           mock.sentinel.qos_policy)
        mock_get_policy_values.assert_called_once_with(
            mock.sentinel.qos_policy)
        self.driver._utils.set_port_qos_rule.assert_called_once_with(
            mock.sentinel.port_id, mock_get_policy_values.return_value)

    @mock.patch.object(qos_driver.QosHyperVAgentDriver, '_get_policy_values')
    def test_update(self, mock_get_policy_values):
        self.driver.update({'port_id': mock.sentinel.port_id},
                           mock.sentinel.qos_policy)
        mock_get_policy_values.assert_called_once_with(
            mock.sentinel.qos_policy)
        self.driver._utils.set_port_qos_rule.assert_called_once_with(
            mock.sentinel.port_id, mock_get_policy_values.return_value)

    def test_delete(self):
        self.driver.delete({'port_id': mock.sentinel.port_id})
        self.driver._utils.remove_port_qos_rule.assert_called_once_with(
            mock.sentinel.port_id)

    def test_get_policy_values(self):
        qos_rule_0 = mock.Mock(spec=['min_kbps', 'rule_type'])
        qos_rule_0.rule_type = qos_consts.RULE_TYPE_MINIMUM_BANDWIDTH
        qos_rule_1 = mock.Mock(spec=['max_kbps', 'max_burst_kbps',
                                     'rule_type'])
        qos_rule_1.rule_type = qos_consts.RULE_TYPE_BANDWIDTH_LIMIT
        qos_policy = mock.Mock(rules=[qos_rule_0, qos_rule_1])

        expected_val = dict(min_kbps=qos_rule_0.min_kbps,
                            max_kbps=qos_rule_1.max_kbps,
                            max_burst_kbps=qos_rule_1.max_burst_kbps)
        policy_val = self.driver._get_policy_values(qos_policy)

        self.assertEqual(expected_val, policy_val)



# Copyright 2013 Cloudbase Solutions SRL
# Copyright 2013 Pedro Navarro Perez
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Unit tests for Windows Hyper-V virtual switch neutron driver
"""

import sys

import ddt
import mock
from neutron.agent import rpc as agent_rpc
from neutron_lib.agent import topics
from neutron_lib import rpc as n_rpc
from os_win import exceptions

from networking_hyperv.neutron.agent import hyperv_neutron_agent as \
    hyperv_agent
from networking_hyperv.neutron.agent import layer2
from networking_hyperv.neutron import constants
from networking_hyperv.neutron import exception
from networking_hyperv.tests import base


class TestHyperVSecurityAgent(base.BaseTestCase):
    @mock.patch.object(hyperv_agent.HyperVSecurityAgent, '__init__',
                       lambda *args, **kwargs: None)
    def setUp(self):
        super(TestHyperVSecurityAgent, self).setUp()
        self.agent = hyperv_agent.HyperVSecurityAgent()

    @mock.patch.object(hyperv_agent, 'HyperVSecurityCallbackMixin')
    @mock.patch.object(hyperv_agent.agent_rpc, 'create_consumers')
    def test_setup_rpc(self, mock_create_consumers, mock_HyperVSecurity):
        self.agent._setup_rpc()

        self.assertEqual(topics.AGENT, self.agent.topic)
        self.assertEqual([mock_HyperVSecurity.return_value],
                         self.agent.endpoints)
        self.assertEqual(mock_create_consumers.return_value,
                         self.agent.connection)
        mock_create_consumers.assert_called_once_with(
            self.agent.endpoints, self.agent.topic,
            [[topics.SECURITY_GROUP, topics.UPDATE]])


class TestHyperVNeutronAgent(base.HyperVBaseTestCase):

    _FAKE_PORT_ID = 'fake_port_id'

    @mock.patch.object(hyperv_agent.HyperVNeutronAgent, "_setup")
    @mock.patch.object(hyperv_agent.HyperVNeutronAgent, "_setup_rpc")
    @mock.patch.object(hyperv_agent.HyperVNeutronAgent, "_set_agent_state")
    def _get_agent(self, mock_set_agent_state, mock_setup_rpc, mock_setup):
        return hyperv_agent.HyperVNeutronAgent()

    def setUp(self):
        super(TestHyperVNeutronAgent, self).setUp()
        self.agent = self._get_agent()

        self.agent._utils = mock.MagicMock(autospec=self.agent._utils)
        self.agent._metricsutils = mock.MagicMock(
            autospec=self.agent._metricsutils)
        self.agent._nvgre_ops = mock.MagicMock(
            autospec=hyperv_agent.nvgre_ops.HyperVNvgreOps)

        self.agent._sec_groups_agent = mock.MagicMock(
            autospec=hyperv_agent.HyperVSecurityAgent)
        self.agent._vlan_driver = mock.MagicMock(
            autospec=hyperv_agent.trunk_driver.HyperVTrunkDriver)
        self.agent._qos_ext = mock.MagicMock(
            autospec=hyperv_agent.qos_extension.QosAgentExtension)

        self.agent._plugin_rpc = mock.MagicMock(autospec=agent_rpc.PluginApi)
        self.agent._client = mock.MagicMock(autospec=n_rpc.BackingOffClient)
        self.agent._connection = mock.MagicMock(autospec=n_rpc.Connection)
        self.agent._refresh_cache = False
        self.agent._added_ports = set()

    def test_get_agent_configurations(self):
        self.agent._physical_network_mappings = mock.sentinel.mappings
        fake_ip = '10.10.10.10'
        self.config(enable_support=True,
                    provider_tunnel_ip=fake_ip,
                    group="NVGRE")

        agent_configurations = self.agent._get_agent_configurations()

        expected_keys = ["vswitch_mappings", "arp_responder_enabled",
                         "tunneling_ip", "devices", "l2_population",
                         "tunnel_types", "enable_distributed_routing",
                         "bridge_mappings"]
        self.assertEqual(sorted(expected_keys),
                         sorted(agent_configurations.keys()))
        self.assertEqual(mock.sentinel.mappings,
                         agent_configurations["vswitch_mappings"])
        self.assertEqual(fake_ip,
                         agent_configurations["tunneling_ip"])

    @mock.patch("networking_hyperv.neutron.trunk_driver.HyperVTrunkDriver")
    @mock.patch("neutron.api.rpc.handlers.securitygroups_rpc."
                "SecurityGroupServerRpcApi")
    @mock.patch("networking_hyperv.neutron.agent.hyperv_neutron_agent"
                ".HyperVSecurityAgent")
    @mock.patch.object(layer2.Layer2Agent, "_setup")
    def test_setup(self, mock_setup, mock_security_agent, mock_sg_rpc,
                   mock_trunk_driver):
        self.agent._context = mock.sentinel.admin_context
        self.agent._consumers = []
        self.config(enable_support=True, group="NVGRE")

        self.agent._setup()

        expected_consumers = [[constants.TUNNEL, topics.UPDATE],
                              [constants.LOOKUP, constants.UPDATE]]
        mock_setup.assert_called_once_with()
        mock_sg_rpc.assert_called_once_with(topics.PLUGIN)
        mock_security_agent.assert_called_once_with(
            mock.sentinel.admin_context, mock_sg_rpc.return_value)
        mock_trunk_driver.assert_called_once_with(mock.sentinel.admin_context)
        self.assertEqual(expected_consumers, self.agent._consumers)

    @mock.patch("neutron.agent.l2.extensions.qos.QosAgentExtension")
    def test_setup_qos_extension(self, mock_qos_agent):
        self.config(enable_qos_extension=True, group="AGENT")
        mock_qos_agent_extension = mock.Mock()
        mock_qos_agent.return_value = mock_qos_agent_extension

        self.agent._setup_qos_extension()

        mock_qos_agent_extension.consume_api.assert_called_once_with(
            self.agent)
        mock_qos_agent_extension.initialize(self.agent._connection, "hyperv")

    @mock.patch.object(hyperv_agent.nvgre_ops, 'HyperVNvgreOps')
    def test_init_nvgre_disabled(self, mock_hyperv_nvgre_ops):
        self.agent._init_nvgre()
        self.assertFalse(mock_hyperv_nvgre_ops.called)
        self.assertFalse(self.agent._nvgre_enabled)

    @mock.patch.object(hyperv_agent.nvgre_ops, 'HyperVNvgreOps')
    def test_init_nvgre_no_tunnel_ip(self, mock_hyperv_nvgre_ops):
        self.config(enable_support=True, group='NVGRE')
        self.assertRaises(exception.NetworkingHyperVException,
                          self.agent._init_nvgre)

    @mock.patch.object(hyperv_agent.nvgre_ops, 'HyperVNvgreOps')
    def test_init_nvgre_enabled(self, mock_hyperv_nvgre_ops):
        self.config(enable_support=True, group='NVGRE')
        fake_ip = '10.10.10.10'
        self.config(provider_tunnel_ip=fake_ip,
                    group='NVGRE')
        self.agent._init_nvgre()
        mock_hyperv_nvgre_ops.assert_called_once_with(
            list(self.agent._physical_network_mappings.values()))

        self.assertTrue(self.agent._nvgre_enabled)
        self.agent._nvgre_ops.init_notifier.assert_called_once_with(
            self.agent._context, self.agent._client)

    @mock.patch.object(hyperv_agent.HyperVNeutronAgent,
                       "_get_vswitch_name")
    def test_provision_network_exception(self, mock_get_vswitch_name):
        self.assertRaises(exception.NetworkingHyperVException,
                          self.agent._provision_network,
                          mock.sentinel.FAKE_PORT_ID,
                          mock.sentinel.FAKE_NET_UUID,
                          mock.sentinel.FAKE_NETWORK_TYPE,
                          mock.sentinel.FAKE_PHYSICAL_NETWORK,
                          mock.sentinel.FAKE_SEGMENTATION_ID)
        mock_get_vswitch_name.assert_called_once_with(
            mock.sentinel.FAKE_NETWORK_TYPE,
            mock.sentinel.FAKE_PHYSICAL_NETWORK)

    @mock.patch.object(hyperv_agent.HyperVNeutronAgent,
                       "_get_vswitch_name")
    def test_provision_network_vlan(self, mock_get_vswitch_name):
        self.agent._provision_network(mock.sentinel.FAKE_PORT_ID,
                                      mock.sentinel.FAKE_NET_UUID,
                                      constants.TYPE_VLAN,
                                      mock.sentinel.FAKE_PHYSICAL_NETWORK,
                                      mock.sentinel.FAKE_SEGMENTATION_ID)
        mock_get_vswitch_name.assert_called_once_with(
            constants.TYPE_VLAN,
            mock.sentinel.FAKE_PHYSICAL_NETWORK)

    @mock.patch.object(hyperv_agent.HyperVNeutronAgent,
                       "_get_vswitch_name")
    def test_provision_network_nvgre(self, mock_get_vswitch_name):
        self.agent._nvgre_enabled = True
        vswitch_name = mock_get_vswitch_name.return_value
        self.agent._provision_network(mock.sentinel.FAKE_PORT_ID,
                                      mock.sentinel.FAKE_NET_UUID,
                                      constants.TYPE_NVGRE,
                                      mock.sentinel.FAKE_PHYSICAL_NETWORK,
                                      mock.sentinel.FAKE_SEGMENTATION_ID)

        mock_get_vswitch_name.assert_called_once_with(
            constants.TYPE_NVGRE,
            mock.sentinel.FAKE_PHYSICAL_NETWORK)
        self.agent._nvgre_ops.bind_nvgre_network.assert_called_once_with(
            mock.sentinel.FAKE_SEGMENTATION_ID,
            mock.sentinel.FAKE_NET_UUID,
            vswitch_name)

    @mock.patch.object(hyperv_agent.HyperVNeutronAgent,
                       "_get_vswitch_name")
    def test_provision_network_flat(self, mock_get_vswitch_name):
        self.agent._provision_network(mock.sentinel.FAKE_PORT_ID,
                                      mock.sentinel.FAKE_NET_UUID,
                                      constants.TYPE_FLAT,
                                      mock.sentinel.FAKE_PHYSICAL_NETWORK,
                                      mock.sentinel.FAKE_SEGMENTATION_ID)
        mock_get_vswitch_name.assert_called_once_with(
            constants.TYPE_FLAT,
            mock.sentinel.FAKE_PHYSICAL_NETWORK)

    @mock.patch.object(hyperv_agent.HyperVNeutronAgent,
                       "_get_vswitch_name")
    def test_provision_network_local(self, mock_get_vswitch_name):
        self.agent._provision_network(mock.sentinel.FAKE_PORT_ID,
                                      mock.sentinel.FAKE_NET_UUID,
                                      constants.TYPE_LOCAL,
                                      mock.sentinel.FAKE_PHYSICAL_NETWORK,
                                      mock.sentinel.FAKE_SEGMENTATION_ID)
        mock_get_vswitch_name.assert_called_once_with(
            constants.TYPE_LOCAL,
            mock.sentinel.FAKE_PHYSICAL_NETWORK)

    @ddt.data({'enable_security_groups': True},
              {'network_type': constants.TYPE_LOCAL},
              {'network_type': constants.TYPE_VLAN},
              {'network_type': constants.TYPE_NVGRE})
    @ddt.unpack
    @mock.patch.object(layer2.Layer2Agent, '_port_bound')
    def test_port_bound(self, mock_super_bound,
                        network_type=constants.TYPE_FLAT,
                        enable_security_groups=False):
        net_uuid = 'my-net-uuid'
        self.agent._enable_metrics_collection = True
        self.agent._enable_security_groups = enable_security_groups
        self.agent._network_vswitch_map[net_uuid] = mock.sentinel.vswitch_name

        self.agent._port_bound(mock.sentinel.port_id, net_uuid,
                               network_type,
                               mock.sentinel.physical_network,
                               mock.sentinel.segmentation_id,
                               mock.sentinel.port_security_enabled,
                               False)

        mock_super_bound.assert_called_once_with(
            mock.sentinel.port_id, net_uuid, network_type,
            mock.sentinel.physical_network,
            mock.sentinel.segmentation_id, mock.sentinel.port_security_enabled,
            False)

        if network_type == constants.TYPE_VLAN:
            self.agent._vlan_driver.bind_vlan_port.assert_called_once_with(
                mock.sentinel.port_id, mock.sentinel.segmentation_id)
        elif network_type == constants.TYPE_NVGRE:
            self.agent._nvgre_ops.bind_nvgre_port.assert_called_once_with(
                mock.sentinel.segmentation_id, mock.sentinel.vswitch_name,
                mock.sentinel.port_id)

        self.agent._utils.add_metrics_collection_acls.assert_called_once_with(
            mock.sentinel.port_id)

        if enable_security_groups:
            refresh_firewall = self.agent._sec_groups_agent.refresh_firewall
            refresh_firewall.assert_called_once_with([mock.sentinel.port_id])
        else:
            remove_sec_rules = self.agent._utils.remove_all_security_rules
            remove_sec_rules.assert_called_once_with(mock.sentinel.port_id)

        set_mac_spoofing = self.agent._utils.set_vswitch_port_mac_spoofing
        set_mac_spoofing.assert_called_once_with(
            mock.sentinel.port_id, mock.sentinel.port_security_enabled)

    def test_port_enable_control_metrics_ok(self):
        self.agent._enable_metrics_collection = True
        self.agent._port_metric_retries[self._FAKE_PORT_ID] = (
            self.agent._metrics_max_retries)

        self.agent._utils.is_metrics_collection_allowed.return_value = True
        self.agent._port_enable_control_metrics()

        enable_port_metrics_collection = (
            self.agent._metricsutils.enable_port_metrics_collection)
        enable_port_metrics_collection.assert_called_with(self._FAKE_PORT_ID)
        self.assertNotIn(self._FAKE_PORT_ID, self.agent._port_metric_retries)

    def test_port_enable_control_metrics_maxed(self):
        self.agent._enable_metrics_collection = True
        self.agent._metrics_max_retries = 3
        self.agent._port_metric_retries[self._FAKE_PORT_ID] = 3

        self.agent._utils.is_metrics_collection_allowed.return_value = False
        for _ in range(4):
            self.assertIn(self._FAKE_PORT_ID,
                          self.agent._port_metric_retries)
            self.agent._port_enable_control_metrics()

        self.assertNotIn(self._FAKE_PORT_ID, self.agent._port_metric_retries)

    def test_port_enable_control_metrics_no_vnic(self):
        self.agent._enable_metrics_collection = True
        self.agent._port_metric_retries[self._FAKE_PORT_ID] = 3
        self.agent._utils.is_metrics_collection_allowed.side_effect = (
            exceptions.NotFound(resource=self._FAKE_PORT_ID))

        self.agent._port_enable_control_metrics()
        self.assertNotIn(self._FAKE_PORT_ID, self.agent._port_metric_retries)

    @mock.patch.object(layer2.Layer2Agent, '_port_unbound')
    def test_port_unbound(self, mock_super_unbound):
        self.agent._port_unbound(mock.sentinel.port_id,
                                 mock.sentinel.vnic_deleted)

        mock_super_unbound.assert_called_once_with(
            mock.sentinel.port_id, mock.sentinel.vnic_deleted)
        remove_dev_filter = self.agent._sec_groups_agent.remove_devices_filter
        remove_dev_filter.assert_called_once_with([mock.sentinel.port_id])

    @mock.patch.object(layer2.Layer2Agent, "_process_added_port")
    def test_process_added_port(self, mock_process):
        self.config(enable_qos_extension=True, group="AGENT")

        self.agent._process_added_port(mock.sentinel.device_details)

        mock_process.assert_called_once_with(mock.sentinel.device_details)
        self.agent._qos_ext.handle_port.assert_called_once_with(
            self.agent._context, mock.sentinel.device_details)

    @mock.patch.object(hyperv_agent.HyperVNeutronAgent,
                       '_port_unbound')
    @mock.patch.object(hyperv_agent.HyperVNeutronAgent,
                       '_update_port_status_cache')
    def test_process_removed_port_exception(self, mock_update_port_cache,
                                            mock_port_unbound):
        self.agent._removed_ports = set([mock.sentinel.port_id])
        remove_devices = self.agent._sec_groups_agent.remove_devices_filter
        remove_devices.side_effect = exception.NetworkingHyperVException

        self.assertRaises(exception.NetworkingHyperVException,
                          self.agent._process_removed_port,
                          mock.sentinel.port_id)

        mock_update_port_cache.assert_called_once_with(
            mock.sentinel.port_id, device_bound=False)
        self.assertIn(mock.sentinel.port_id, self.agent._removed_ports)

    @mock.patch.object(hyperv_agent.HyperVNeutronAgent,
                       '_port_unbound')
    @mock.patch.object(hyperv_agent.HyperVNeutronAgent,
                       '_update_port_status_cache')
    def test_process_removed_port(self, mock_update_port_cache,
                                  mock_port_unbound):
        self.agent._removed_ports = set([mock.sentinel.port_id])

        self.agent._process_removed_port(mock.sentinel.port_id)

        mock_update_port_cache.assert_called_once_with(
            mock.sentinel.port_id, device_bound=False)
        mock_port_unbound.assert_called_once_with(mock.sentinel.port_id,
                                                  vnic_deleted=True)
        self.agent._sec_groups_agent.remove_devices_filter(
            [mock.sentinel.port_id])
        self.assertNotIn(mock.sentinel.port_id, self.agent._removed_ports)

    @mock.patch.object(layer2.Layer2Agent, "_work")
    @mock.patch.object(hyperv_agent.HyperVNeutronAgent,
                       '_port_enable_control_metrics')
    def test_work(self, mock_port_enable_metrics, mock_work):
        self.agent._nvgre_enabled = True

        self.agent._work()

        mock_work.assert_called_once_with()
        self.agent._nvgre_ops.refresh_nvgre_records.assert_called_once_with()
        mock_port_enable_metrics.assert_called_with()


class TestMain(base.BaseTestCase):

    @mock.patch.object(hyperv_agent, 'HyperVNeutronAgent')
    @mock.patch.object(hyperv_agent, 'common_config')
    @mock.patch.object(hyperv_agent, 'neutron_config')
    def test_main(self, mock_config, mock_common_config, mock_hyperv_agent):
        hyperv_agent.main()

        mock_config.register_agent_state_opts_helper.assert_called_once_with(
            hyperv_agent.CONF)
        mock_common_config.init.assert_called_once_with(sys.argv[1:])
        mock_config.setup_logging.assert_called_once_with()
        mock_hyperv_agent.assert_called_once_with()
        mock_hyperv_agent.return_value.daemon_loop.assert_called_once_with()



# Copyright 2013 Cloudbase Solutions SRL
# Copyright 2013 Pedro Navarro Perez
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import platform
import sys

from neutron.agent.l2.extensions import qos as qos_extension
from neutron.agent import rpc as agent_rpc
from neutron.agent import securitygroups_rpc as sg_agent_rpc
from neutron.api.rpc.handlers import securitygroups_rpc as sg_rpc
from neutron.common import config as common_config
from neutron.conf.agent import common as neutron_config
from neutron_lib.agent import topics
from os_win import exceptions
from os_win import utilsfactory
from oslo_log import log as logging
import oslo_messaging

from networking_hyperv.common.i18n import _, _LI, _LW, _LE    # noqa
from networking_hyperv.neutron import _common_utils as c_util
from networking_hyperv.neutron.agent import layer2 as hyperv_base
from networking_hyperv.neutron import config
from networking_hyperv.neutron import constants as h_constant
from networking_hyperv.neutron import exception
from networking_hyperv.neutron import nvgre_ops
from networking_hyperv.neutron import trunk_driver

CONF = config.CONF
LOG = logging.getLogger(__name__)

_port_synchronized = c_util.get_port_synchronized_decorator('n-hv-agent-')


class HyperVSecurityAgent(sg_agent_rpc.SecurityGroupAgentRpc):

    def __init__(self, context, plugin_rpc):
        super(HyperVSecurityAgent, self).__init__(context, plugin_rpc)
        if sg_agent_rpc.is_firewall_enabled():
            self._setup_rpc()

    @property
    def use_enhanced_rpc(self):
        return True

    def _setup_rpc(self):
        self.topic = topics.AGENT
        self.endpoints = [HyperVSecurityCallbackMixin(self)]
        consumers = [[topics.SECURITY_GROUP, topics.UPDATE]]

        self.connection = agent_rpc.create_consumers(self.endpoints,
                                                     self.topic,
                                                     consumers)


class HyperVSecurityCallbackMixin(sg_rpc.SecurityGroupAgentRpcCallbackMixin):

    target = oslo_messaging.Target(version='1.3')

    def __init__(self, sg_agent):
        super(HyperVSecurityCallbackMixin, self).__init__()
        self.sg_agent = sg_agent


class HyperVNeutronAgent(hyperv_base.Layer2Agent):

    _AGENT_BINARY = "neutron-hyperv-agent"
    _AGENT_TYPE = h_constant.AGENT_TYPE_HYPERV

    def __init__(self):
        super(HyperVNeutronAgent, self).__init__()
        self._agent_id = 'hyperv_%s' % platform.node()

        self._qos_ext = None
        self._nvgre_enabled = False

        self._metricsutils = utilsfactory.get_metricsutils()
        self._port_metric_retries = {}

        agent_conf = CONF.get('AGENT', {})
        security_conf = CONF.get('SECURITYGROUP', {})
        self._enable_metrics_collection = agent_conf.get(
            'enable_metrics_collection', False)
        self._metrics_max_retries = agent_conf.get('metrics_max_retries', 100)
        self._enable_security_groups = security_conf.get(
            'enable_security_group', False)

        self._init_nvgre()

    def _get_agent_configurations(self):
        configurations = {'vswitch_mappings': self._physical_network_mappings}
        if CONF.NVGRE.enable_support:
            configurations['arp_responder_enabled'] = False
            configurations['tunneling_ip'] = CONF.NVGRE.provider_tunnel_ip
            configurations['devices'] = 1
            configurations['l2_population'] = False
            configurations['tunnel_types'] = [h_constant.TYPE_NVGRE]
            configurations['enable_distributed_routing'] = False
            configurations['bridge_mappings'] = {}
        return configurations

    def _setup(self):
        """Setup the layer two agent."""
        super(HyperVNeutronAgent, self)._setup()

        self._sg_plugin_rpc = sg_rpc.SecurityGroupServerRpcApi(topics.PLUGIN)
        self._sec_groups_agent = HyperVSecurityAgent(self._context,
                                                     self._sg_plugin_rpc)
        self._vlan_driver = trunk_driver.HyperVTrunkDriver(self._context)
        if CONF.NVGRE.enable_support:
            self._consumers.append([h_constant.TUNNEL, topics.UPDATE])
            self._consumers.append([h_constant.LOOKUP, h_constant.UPDATE])

    def _setup_qos_extension(self):
        """Setup the QOS extension if it is required."""
        if not CONF.AGENT.enable_qos_extension:
            return
        self._qos_ext = qos_extension.QosAgentExtension()
        self._qos_ext.consume_api(self)
        self._qos_ext.initialize(self._connection, 'hyperv')

    def _init_nvgre(self):
        # if NVGRE is enabled, self._nvgre_ops is required in order to properly
        # set the agent state (see get_agent_configurations method).
        if not CONF.NVGRE.enable_support:
            return

        if not CONF.NVGRE.provider_tunnel_ip:
            err_msg = _('enable_nvgre_support is set to True, but '
                        'provider tunnel IP is not configured. '
                        'Check neutron.conf config file.')
            LOG.error(err_msg)
            raise exception.NetworkingHyperVException(err_msg)

        self._nvgre_enabled = True
        self._nvgre_ops = nvgre_ops.HyperVNvgreOps(
            list(self._physical_network_mappings.values()))

        self._nvgre_ops.init_notifier(self._context, self._client)
        self._nvgre_ops.tunnel_update(self._context,
                                      CONF.NVGRE.provider_tunnel_ip,
                                      h_constant.TYPE_NVGRE)

    def _provision_network(self, port_id, net_uuid, network_type,
                           physical_network, segmentation_id):
        """Provision the network with the received information."""
        LOG.info("Provisioning network %s", net_uuid)

        vswitch_name = self._get_vswitch_name(network_type, physical_network)
        if network_type == h_constant.TYPE_VLAN:
            # Nothing to do
            pass
        elif network_type == h_constant.TYPE_FLAT:
            # Nothing to do
            pass
        elif network_type == h_constant.TYPE_LOCAL:
            # TODO(alexpilotti): Check that the switch type is private
            # or create it if not existing.
            pass
        elif network_type == h_constant.TYPE_NVGRE and self._nvgre_enabled:
            self._nvgre_ops.bind_nvgre_network(segmentation_id, net_uuid,
                                               vswitch_name)
        else:
            raise exception.NetworkingHyperVException(
                (_("Cannot provision unknown network type "
                   "%(network_type)s for network %(net_uuid)s") %
                 dict(network_type=network_type, net_uuid=net_uuid)))

        vswitch_map = {
            'network_type': network_type,
            'vswitch_name': vswitch_name,
            'ports': [],
            'vlan_id': segmentation_id}
        self._network_vswitch_map[net_uuid] = vswitch_map

    def _port_bound(self, port_id, network_id, network_type, physical_network,
                    segmentation_id, port_security_enabled, set_port_sriov):
        """Bind the port to the recived network."""
        super(HyperVNeutronAgent, self)._port_bound(
            port_id, network_id, network_type, physical_network,
            segmentation_id, port_security_enabled, set_port_sriov
        )
        vswitch_map = self._network_vswitch_map[network_id]

        if network_type == h_constant.TYPE_VLAN:
            self._vlan_driver.bind_vlan_port(port_id, segmentation_id)
        elif network_type == h_constant.TYPE_NVGRE and self._nvgre_enabled:
            self._nvgre_ops.bind_nvgre_port(
                segmentation_id, vswitch_map['vswitch_name'], port_id)
        elif network_type == h_constant.TYPE_FLAT:
            pass    # Nothing to do
        elif network_type == h_constant.TYPE_LOCAL:
            pass    # Nothing to do
        else:
            LOG.error('Unsupported network type %s', network_type)

        if self._enable_metrics_collection:
            self._utils.add_metrics_collection_acls(port_id)
            self._port_metric_retries[port_id] = self._metrics_max_retries

        # check if security groups is enabled.
        # if not, teardown the security group rules
        if self._enable_security_groups:
            self._sec_groups_agent.refresh_firewall([port_id])
        else:
            self._utils.remove_all_security_rules(port_id)

        self._utils.set_vswitch_port_mac_spoofing(port_id,
                                                  port_security_enabled)

    def _port_enable_control_metrics(self):
        if not self._enable_metrics_collection:
            return

        for port_id in list(self._port_metric_retries.keys()):
            try:
                if self._utils.is_metrics_collection_allowed(port_id):
                    self._metricsutils.enable_port_metrics_collection(port_id)
                    LOG.info('Port metrics enabled for port: %s', port_id)
                    del self._port_metric_retries[port_id]
                elif self._port_metric_retries[port_id] < 1:
                    self._metricsutils.enable_port_metrics_collection(port_id)
                    LOG.error('Port metrics raw enabling for port: %s',
                              port_id)
                    del self._port_metric_retries[port_id]
                else:
                    self._port_metric_retries[port_id] -= 1
            except exceptions.NotFound:
                # the vNIC no longer exists. it might have been removed or
                # the VM it was attached to was destroyed.
                LOG.warning("Port %s no longer exists. Cannot enable "
                            "metrics.", port_id)
                del self._port_metric_retries[port_id]

    def _port_unbound(self, port_id, vnic_deleted=False):
        super(HyperVNeutronAgent, self)._port_unbound(port_id, vnic_deleted)
        self._sec_groups_agent.remove_devices_filter([port_id])

    def _process_added_port(self, device_details):
        super(HyperVNeutronAgent, self)._process_added_port(
            device_details)

        if CONF.AGENT.enable_qos_extension:
            self._qos_ext.handle_port(self._context, device_details)

    def _process_removed_port(self, device):
        super(HyperVNeutronAgent, self)._process_removed_port(device)
        try:
            self._sec_groups_agent.remove_devices_filter([device])
        except Exception:
            LOG.exception("Exception encountered while processing"
                          " port %s.", device)
            # Readd the port as "removed", so it can be reprocessed.
            self._removed_ports.add(device)
            raise

    def _work(self):
        """Process the information regarding the available ports."""
        super(HyperVNeutronAgent, self)._work()
        if self._nvgre_enabled:
            self._nvgre_ops.refresh_nvgre_records()
        self._port_enable_control_metrics()

    def tunnel_update(self, context, **kwargs):
        LOG.info('tunnel_update received: kwargs: %s', kwargs)
        tunnel_ip = kwargs.get('tunnel_ip')
        if tunnel_ip == CONF.NVGRE.provider_tunnel_ip:
            # the notification should be ignored if it originates from this
            # node.
            return

        tunnel_type = kwargs.get('tunnel_type')
        self._nvgre_ops.tunnel_update(context, tunnel_ip, tunnel_type)

    def lookup_update(self, context, **kwargs):
        self._nvgre_ops.lookup_update(kwargs)


def main():
    """The entry point for the Hyper-V Neutron Agent."""
    neutron_config.register_agent_state_opts_helper(CONF)
    common_config.init(sys.argv[1:])
    neutron_config.setup_logging()

    hyperv_agent = HyperVNeutronAgent()

    # Start everything.
    LOG.info("Agent initialized successfully, now running... ")
    hyperv_agent.daemon_loop()



# Copyright 2013 Cloudbase Solutions SRL
# Copyright 2013 Pedro Navarro Perez
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ddt
import mock
from os_win import exceptions as os_win_exc

from networking_hyperv.neutron import _common_utils
from networking_hyperv.tests import base


@ddt.ddt
class TestCommonUtils(base.BaseTestCase):

    @mock.patch.object(_common_utils.lockutils, 'synchronized_with_prefix')
    def test_create_synchronized_decorator(self, mock_sync_with_prefix):
        fake_method_side_effect = mock.Mock()
        lock_prefix = 'test-'
        port_synchronized = _common_utils.get_port_synchronized_decorator(
            lock_prefix)

        @port_synchronized
        def fake_method(fake_arg, port_id):
            fake_method_side_effect(fake_arg, port_id)

        mock_synchronized = mock_sync_with_prefix.return_value
        mock_synchronized.return_value = lambda x: x
        expected_lock_name = 'test-port-lock-%s' % mock.sentinel.port_id

        fake_method(fake_arg=mock.sentinel.arg, port_id=mock.sentinel.port_id)
        mock_sync_with_prefix.assert_called_once_with(lock_prefix)
        mock_synchronized.assert_called_once_with(expected_lock_name)
        fake_method_side_effect.assert_called_once_with(
            mock.sentinel.arg, mock.sentinel.port_id)

    @ddt.data(os_win_exc.HyperVPortNotFoundException(message='test'),
              os_win_exc.HyperVvNicNotFound(message='test'))
    def test_ignore_missing_ports_decorator(self, exc):
        fake_method_side_effect = mock.Mock()
        fake_method_side_effect.side_effect = exc

        @_common_utils.ignore_missing_ports
        def fake_method(fake_arg, port_id):
            fake_method_side_effect(fake_arg, port_id)

        fake_method(mock.sentinel.arg, mock.sentinel.port_id)

        fake_method_side_effect.assert_called_once_with(
            mock.sentinel.arg, mock.sentinel.port_id)



# Copyright 2017 Cloudbase Solutions SRL
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""This module contains the contract class for each agent."""

import abc
import threading
import time

from neutron_lib.agent import topics
from neutron_lib import context as neutron_context
from os_win import utilsfactory
from oslo_log import log as logging
import oslo_messaging
import six

from networking_hyperv.common.i18n import _LE    # noqa
from networking_hyperv.neutron import config


LOG = logging.getLogger(__name__)
CONF = config.CONF


@six.add_metaclass(abc.ABCMeta)
class BaseAgent(object):

    """Contact class for all the neutron agents."""

    _AGENT_BINARY = None
    _AGENT_TYPE = None
    _AGENT_TOPIC = None

    target = oslo_messaging.Target(version='1.3')

    def __init__(self):
        """Initializes local configuration of the current agent.

        :param conf: dict or dict-like object containing the configuration
                     details used by this Agent. If None is specified, default
                     values are used instead.
        """
        self._agent_id = None
        self._topic = topics.AGENT
        self._cache_lock = threading.Lock()
        self._refresh_cache = False
        self._host = CONF.get("host")

        self._agent_state = {}
        self._context = neutron_context.get_admin_context_without_session()

        self._utils = utilsfactory.get_networkutils()
        self._utils.init_caches()

        # The following attributes will be initialized by the
        # `_setup_rpc` method.
        self._client = None
        self._connection = None
        self._endpoints = []
        self._plugin_rpc = None
        self._sg_plugin_rpc = None
        self._state_rpc = None

        agent_config = CONF.get("AGENT", {})
        self._polling_interval = agent_config.get('polling_interval', 2)

    @abc.abstractmethod
    def _get_agent_configurations(self):
        """Get configurations for the current agent."""
        pass

    def _set_agent_state(self):
        """Set the state for the agent."""
        self._agent_state = {
            'agent_type': self._AGENT_TYPE,
            'binary': self._AGENT_BINARY,
            'configurations': self._get_agent_configurations(),
            'host': self._host,
            'start_flag': True,
            'topic': self._AGENT_TOPIC,
        }

    @abc.abstractmethod
    def _setup_rpc(self):
        """Setup the RPC client for the current agent."""
        pass

    @abc.abstractmethod
    def _work(self):
        """Override this with your desired procedures."""
        pass

    def _prologue(self):
        """Executed once before the daemon loop."""
        pass

    def daemon_loop(self):
        """Process all the available ports."""
        self._prologue()
        while True:
            start = time.time()
            try:
                self._work()
            except Exception:
                LOG.exception("Error in agent event loop")

            # Sleep until the end of polling interval
            elapsed = (time.time() - start)
            if elapsed < self._polling_interval:
                time.sleep(self._polling_interval - elapsed)
            else:
                LOG.debug("Loop iteration exceeded interval "
                          "(%(polling_interval)s vs. %(elapsed)s)",
                          {'polling_interval': self._polling_interval,
                           'elapsed': elapsed})

    def _report_state(self):
        try:
            self._state_rpc.report_state(self._context,
                                         self._agent_state)
            self._agent_state.pop('start_flag', None)
        except Exception:
            LOG.exception("Failed reporting state!")



# Copyright 2015 Cloudbase Solutions SRL
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Unit tests for Windows Hyper-V NVGRE driver.
"""

import mock

from networking_hyperv.neutron import config
from networking_hyperv.neutron import constants
from networking_hyperv.neutron import nvgre_ops
from networking_hyperv.tests import base

CONF = config.CONF


class TestHyperVNvgreOps(base.HyperVBaseTestCase):

    _autospec_classes = [
        nvgre_ops.neutron_client.NeutronAPIClient,
    ]

    FAKE_MAC_ADDR = 'fa:ke:ma:ca:dd:re:ss'
    FAKE_CIDR = '10.0.0.0/24'
    FAKE_VSWITCH_NAME = 'fake_vswitch'

    def setUp(self):
        super(TestHyperVNvgreOps, self).setUp()

        self.context = 'context'
        self.ops = nvgre_ops.HyperVNvgreOps([])
        self.ops._vswitch_ips[mock.sentinel.network_name] = (
            mock.sentinel.ip_addr)
        self.ops.context = self.context
        self.ops._notifier = mock.MagicMock(
            autospec=nvgre_ops.hyperv_agent_notifier.AgentNotifierApi)
        self.ops._hyperv_utils = mock.MagicMock(
            autospec=self.ops._hyperv_utils)
        self.ops._nvgre_utils = mock.MagicMock(autospec=self.ops._nvgre_utils)

    @mock.patch.object(nvgre_ops.hyperv_agent_notifier, 'AgentNotifierApi')
    def test_init_notifier(self, mock_notifier):
        self.ops.init_notifier(mock.sentinel.context, mock.sentinel.rpc_client)
        mock_notifier.assert_called_once_with(
            constants.AGENT_TOPIC,
            mock.sentinel.rpc_client)
        self.assertEqual(mock_notifier.return_value, self.ops._notifier)
        self.assertEqual(mock.sentinel.context, self.ops.context)

    def test_init_nvgre(self):
        self.ops._nvgre_utils.get_network_iface_ip.return_value = (
            mock.sentinel.ip_addr, mock.sentinel.length)

        self.ops._init_nvgre([mock.sentinel.physical_network])

        self.assertEqual(self.ops._vswitch_ips[mock.sentinel.physical_network],
                         mock.sentinel.ip_addr)
        self.ops._nvgre_utils.create_provider_route.assert_called_once_with(
            mock.sentinel.physical_network)
        self.ops._nvgre_utils.create_provider_address.assert_called_once_with(
            mock.sentinel.physical_network, CONF.NVGRE.provider_vlan_id)

    def test_refresh_tunneling_agents(self):
        self.ops._n_client.get_tunneling_agents.return_value = {
            mock.sentinel.host: mock.sentinel.host_ip
        }
        self.ops._refresh_tunneling_agents()
        self.assertEqual(mock.sentinel.host_ip,
                         self.ops._tunneling_agents[mock.sentinel.host])

    @mock.patch.object(nvgre_ops.HyperVNvgreOps, '_register_lookup_record')
    def test_lookup_update(self, mock_register_record):
        args = {'lookup_ip': mock.sentinel.lookup_ip,
                'lookup_details': {
                    'customer_addr': mock.sentinel.customer_addr,
                    'mac_addr': mock.sentinel.mac_addr,
                    'customer_vsid': mock.sentinel.vsid}
                }

        self.ops.lookup_update(args)

        mock_register_record.assert_called_once_with(
            mock.sentinel.lookup_ip,
            mock.sentinel.customer_addr,
            mock.sentinel.mac_addr,
            mock.sentinel.vsid)

    def test_tunnel_update_nvgre(self):
        self.ops.tunnel_update(
            mock.sentinel.context,
            mock.sentinel.tunnel_ip,
            tunnel_type=constants.TYPE_NVGRE)

        self.ops._notifier.tunnel_update.assert_called_once_with(
            mock.sentinel.context,
            CONF.NVGRE.provider_tunnel_ip,
            constants.TYPE_NVGRE)

    def test_tunnel_update(self):
        self.ops.tunnel_update(
            mock.sentinel.context,
            mock.sentinel.tunnel_ip,
            mock.sentinel.tunnel_type)

        self.assertFalse(self.ops._notifier.tunnel_update.called)

    @mock.patch.object(nvgre_ops.HyperVNvgreOps, '_register_lookup_record')
    def test_lookup_update_no_details(self, mock_register_record):
        self.ops.lookup_update({})
        self.assertFalse(mock_register_record.called)

    def test_register_lookup_record(self):
        self.ops._register_lookup_record(
            mock.sentinel.provider_addr, mock.sentinel.customer_addr,
            mock.sentinel.mac_addr, mock.sentinel.vsid)

        self.ops._nvgre_utils.create_lookup_record.assert_called_once_with(
            mock.sentinel.provider_addr, mock.sentinel.customer_addr,
            mock.sentinel.mac_addr, mock.sentinel.vsid)

    @mock.patch.object(nvgre_ops.HyperVNvgreOps, '_register_lookup_record')
    def test_bind_nvgre_port(self, mock_register_record):
        self.ops._nvgre_utils.get_network_iface_ip.return_value = (
            mock.sentinel.provider_addr, mock.sentinel.prefix_len)

        mac_addr = self.ops._hyperv_utils.get_vnic_mac_address.return_value
        customer_addr = self.ops._n_client.get_port_ip_address.return_value

        self.ops.bind_nvgre_port(mock.sentinel.vsid,
                                 mock.sentinel.network_name,
                                 mock.sentinel.port_id)

        self.ops._hyperv_utils.set_vswitch_port_vsid.assert_called_once_with(
            mock.sentinel.vsid, mock.sentinel.port_id)
        mock_register_record.assert_has_calls([
            mock.call(mock.sentinel.provider_addr, customer_addr, mac_addr,
                      mock.sentinel.vsid),
            mock.call(mock.sentinel.ip_addr, constants.IPV4_DEFAULT, mac_addr,
                      mock.sentinel.vsid)])
        self.ops._notifier.lookup_update.assert_called_once_with(
            self.context, mock.sentinel.provider_addr, {
                'customer_addr': customer_addr,
                'mac_addr': mac_addr,
                'customer_vsid': mock.sentinel.vsid
            })

    def test_bind_nvgre_port_no_provider_addr(self):
        self.ops._nvgre_utils.get_network_iface_ip = mock.MagicMock(
            return_value=(None, None))

        self.ops.bind_nvgre_port(mock.sentinel.vsid,
                                 mock.sentinel.network_name,
                                 mock.sentinel.port_id)

        self.assertFalse(self.ops._hyperv_utils.set_vswitch_port_vsid.called)

    @mock.patch.object(nvgre_ops.HyperVNvgreOps, 'refresh_nvgre_records')
    @mock.patch.object(nvgre_ops.HyperVNvgreOps, '_create_customer_routes')
    def test_bind_nvgre_network(self, mock_create_routes,
                                mock_refresh_records):
        fake_ip = '10.10.10.10'
        self.config(provider_tunnel_ip=fake_ip, group='NVGRE')
        self.ops._n_client.get_network_subnets.return_value = [
            mock.sentinel.subnet, mock.sentinel.subnet2]

        get_cidr = self.ops._n_client.get_network_subnet_cidr_and_gateway
        get_cidr.return_value = (self.FAKE_CIDR, mock.sentinel.gateway)

        self.ops.bind_nvgre_network(
            mock.sentinel.vsid, mock.sentinel.net_uuid,
            self.FAKE_VSWITCH_NAME)

        self.assertEqual(mock.sentinel.vsid,
                         self.ops._network_vsids[mock.sentinel.net_uuid])
        self.ops._n_client.get_network_subnets.assert_called_once_with(
            mock.sentinel.net_uuid)
        get_cidr.assert_called_once_with(mock.sentinel.subnet)
        mock_create_routes.assert_called_once_with(
            mock.sentinel.vsid, self.FAKE_CIDR,
            mock.sentinel.gateway, mock.ANY)
        mock_refresh_records.assert_called_once_with(
            network_id=mock.sentinel.net_uuid)
        self.ops._notifier.tunnel_update.assert_called_once_with(
            self.context, fake_ip, mock.sentinel.vsid)

    def _check_create_customer_routes(self, gateway=None):
        self.ops._create_customer_routes(
            mock.sentinel.vsid, mock.sentinel.cidr,
            gateway, mock.sentinel.rdid)

        self.ops._nvgre_utils.clear_customer_routes.assert_called_once_with(
            mock.sentinel.vsid)
        self.ops._nvgre_utils.create_customer_route.assert_called_once_with(
            mock.sentinel.vsid, mock.sentinel.cidr, constants.IPV4_DEFAULT,
            mock.sentinel.rdid)

    def test_create_customer_routes_no_gw(self):
        self._check_create_customer_routes()

    def test_create_customer_routes_bad_gw(self):
        gateway = '10.0.0.1'
        self._check_create_customer_routes(gateway=gateway)

    def test_create_customer_routes(self):
        gateway = '10.0.0.2'
        self.ops._create_customer_routes(
            mock.sentinel.vsid, mock.sentinel.cidr,
            gateway, mock.sentinel.rdid)

        metadata_addr = '%s/32' % CONF.AGENT.neutron_metadata_address
        self.ops._nvgre_utils.create_customer_route.assert_has_calls([
            mock.call(mock.sentinel.vsid, mock.sentinel.cidr,
                      constants.IPV4_DEFAULT, mock.sentinel.rdid),
            mock.call(mock.sentinel.vsid, '%s/0' % constants.IPV4_DEFAULT,
                      gateway, mock.ANY),
            mock.call(mock.sentinel.vsid, metadata_addr,
                      gateway, mock.ANY)], any_order=True)

    @mock.patch.object(nvgre_ops.HyperVNvgreOps, '_register_lookup_record')
    def test_refresh_nvgre_records(self, mock_register_record):
        self.ops._n_client.get_tunneling_agents.return_value = {}
        self.ops._nvgre_ports.append(mock.sentinel.processed_port_id)
        self.ops._tunneling_agents[mock.sentinel.host_id] = (
            mock.sentinel.agent_ip)
        self.ops._network_vsids[mock.sentinel.net_id] = (
            mock.sentinel.vsid)

        processed_port = {'id': mock.sentinel.processed_port_id}
        no_host_port = {'id': mock.sentinel.port_no_host_id,
                        'binding:host_id': mock.sentinel.odd_host_id}
        other_net_id_port = {'id': mock.sentinel.port_other_net_id,
                             'binding:host_id': mock.sentinel.host_id,
                             'network_id': mock.sentinel.odd_net_id}
        port = {'id': mock.sentinel.port_id,
                'binding:host_id': mock.sentinel.host_id,
                'network_id': mock.sentinel.net_id,
                'mac_address': self.FAKE_MAC_ADDR,
                'fixed_ips': [{'ip_address': mock.sentinel.customer_addr}]
                }

        self.ops._n_client.get_network_ports.return_value = [
            processed_port, no_host_port, other_net_id_port, port]

        self.ops.refresh_nvgre_records()

        expected_mac = self.FAKE_MAC_ADDR.replace(':', '')
        mock_register_record.assert_has_calls([
            mock.call(mock.sentinel.agent_ip, mock.sentinel.customer_addr,
                      expected_mac, mock.sentinel.vsid),
            # mock.call(mock.sentinel.agent_ip, constants.METADATA_ADDR,
            #          expected_mac, mock.sentinel.vsid)
        ])
        self.assertIn(mock.sentinel.port_id, self.ops._nvgre_ports)

    @mock.patch.object(nvgre_ops.HyperVNvgreOps, '_register_lookup_record')
    def test_refresh_nvgre_records_exception(self, mock_register_record):
        self.ops._n_client.get_tunneling_agents.return_value = {}
        self.ops._tunneling_agents[mock.sentinel.host_id] = (
            mock.sentinel.agent_ip)
        self.ops._network_vsids[mock.sentinel.net_id] = (mock.sentinel.vsid)
        port = mock.MagicMock()
        self.ops._n_client.get_network_ports.return_value = [port]
        mock_register_record.side_effect = TypeError

        self.ops.refresh_nvgre_records()

        self.assertNotIn(mock.sentinel.port_id, self.ops._nvgre_ports)



# Copyright 2017 Cloudbase Solutions Srl
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Unit tests for Neutron layer 2 agent.
"""
import collections
import eventlet
import mock

import ddt
import neutron
from neutron.conf.agent import common as neutron_config
from neutron_lib.agent import topics
from neutron_lib import rpc as n_rpc
from os_win import constants as os_win_const
from os_win import exceptions as os_win_exc

from networking_hyperv.neutron.agent import layer2 as agent_base
from networking_hyperv.neutron import config
from networking_hyperv.neutron import constants
from networking_hyperv.neutron import exception
from networking_hyperv.tests import base as test_base

CONF = config.CONF


class _Layer2Agent(agent_base.Layer2Agent):

    def _get_agent_configurations(self):
        pass

    def _report_state(self):
        pass

    def _provision_network(self, port_id, net_uuid, network_type,
                           physical_network, segmentation_id):
        pass


@ddt.ddt
class TestLayer2Agent(test_base.HyperVBaseTestCase):

    _FAKE_PORT_ID = 'fake_port_id'

    @mock.patch.object(_Layer2Agent, "_setup")
    @mock.patch.object(_Layer2Agent, "_setup_rpc")
    @mock.patch.object(_Layer2Agent, "_set_agent_state")
    def _get_agent(self, mock_set_agent_state, mock_setup_rpc, mock_setup):
        return _Layer2Agent()

    def setUp(self):
        super(TestLayer2Agent, self).setUp()
        neutron_config.register_agent_state_opts_helper(CONF)

        self._agent = self._get_agent()

        self._agent._utils = mock.MagicMock(
            autospec=self._agent._utils)
        self._agent._plugin_rpc = mock.Mock(
            autospec=agent_base.agent_rpc.PluginApi)
        self._agent._endpoints = mock.MagicMock()
        self._agent._client = mock.MagicMock(
            autospec=agent_base.n_rpc.BackingOffClient)
        self._agent._connection = mock.MagicMock(
            autospec=agent_base.n_rpc.Connection)
        self._agent._physical_network_mappings = collections.OrderedDict()
        self._agent._network_vswitch_map = {}

    def _get_fake_port_details(self):
        return {
            'device': mock.sentinel.device,
            'port_id': mock.sentinel.port_id,
            'network_id': mock.sentinel.network_id,
            'network_type': mock.sentinel.network_type,
            'physical_network': mock.sentinel.physical_network,
            'segmentation_id': mock.sentinel.segmentation_id,
            'admin_state_up': mock.sentinel.admin_state_up,
            'port_security_enabled': mock.sentinel.port_security_enabled,
        }

    @mock.patch.object(agent_base.Layer2Agent, '_process_removed_port_event',
                       mock.sentinel._process_removed_port_event)
    @mock.patch.object(agent_base.Layer2Agent, '_process_added_port_event',
                       mock.sentinel._process_added_port_event)
    @mock.patch.object(agent_base.Layer2Agent, '_validate_vswitches')
    @mock.patch.object(eventlet.tpool, 'set_num_threads')
    @mock.patch.object(agent_base.Layer2Agent,
                       '_load_physical_network_mappings')
    def test_setup(self, mock_load_phys_net_mapp,
                   mock_set_num_threads,
                   mock_validate_vswitches):
        self.config(
            group="AGENT",
            worker_count=12,
            physical_network_vswitch_mappings=["fake_mappings"],
            local_network_vswitch="local_network_vswitch")
        self._agent._event_callback_pairs = []

        self._agent._setup()

        mock_load_phys_net_mapp.assert_called_once_with(["fake_mappings"])
        mock_validate_vswitches.assert_called_once_with()
        self._agent._endpoints.append.assert_called_once_with(self._agent)
        self.assertIn((self._agent._utils.EVENT_TYPE_CREATE,
                       mock.sentinel._process_added_port_event),
                      self._agent._event_callback_pairs)
        self.assertIn((self._agent._utils.EVENT_TYPE_DELETE,
                       mock.sentinel._process_removed_port_event),
                      self._agent._event_callback_pairs)

    @mock.patch('oslo_service.loopingcall.FixedIntervalLoopingCall')
    @mock.patch.object(agent_base.Layer2Agent, '_setup_qos_extension')
    @mock.patch.object(neutron.agent.rpc, 'create_consumers')
    @mock.patch.object(n_rpc, 'get_client')
    @mock.patch.object(neutron.agent.rpc, 'PluginReportStateAPI')
    @mock.patch.object(neutron.agent.rpc, 'PluginApi')
    def test_setup_rpc(self, mock_plugin_api, mock_plugin_report_state_api,
                       mock_get_client, mock_create_consumers,
                       mock_setup_qos_extension, mock_looping_call):
        self.config(group="AGENT",
                    report_interval=1)
        consumers = [[topics.PORT, topics.UPDATE],
                     [topics.NETWORK, topics.DELETE],
                     [topics.PORT, topics.DELETE]]

        mock_heartbeat = mock.MagicMock()
        mock_create_consumers.return_value = self._agent._connection
        mock_looping_call.return_value = mock_heartbeat

        self._agent._setup_rpc()

        mock_plugin_api.assert_called_once_with(topics.PLUGIN)
        mock_plugin_report_state_api.assert_called_once_with(topics.PLUGIN)
        mock_get_client.assert_called_once_with(self._agent.target)
        self.assertEqual(self._agent._consumers, consumers)
        mock_create_consumers.assert_called_once_with(
            self._agent._endpoints, self._agent._topic, self._agent._consumers,
            start_listening=False)
        mock_setup_qos_extension.assert_called_once_with()
        self._agent._connection.consume_in_threads.assert_called_once_with()
        mock_looping_call.assert_called_once_with(self._agent._report_state)
        mock_heartbeat.start.assert_called_once_with(
            interval=CONF.AGENT.report_interval)

    def test_process_added_port_event(self):
        self._agent._added_ports = set()
        self._agent._process_added_port_event(mock.sentinel.port_id)
        self.assertIn(mock.sentinel.port_id, self._agent._added_ports)

    def test_process_removed_port_event(self):
        self._agent._removed_ports = set([])
        self._agent._process_removed_port_event(mock.sentinel.port_id)
        self.assertIn(mock.sentinel.port_id, self._agent._removed_ports)

    def test_load_physical_network_mappings(self):
        test_mappings = [
            'fakenetwork1:fake_vswitch', 'fakenetwork2:fake_vswitch_2',
            '*:fake_vswitch_3', 'bad_mapping'
        ]
        expected = [
            ('fakenetwork1$', 'fake_vswitch'),
            ('fakenetwork2$', 'fake_vswitch_2'),
            ('.*$', 'fake_vswitch_3')
        ]
        self._agent._physical_network_mappings = collections.OrderedDict()

        self._agent._load_physical_network_mappings(test_mappings)

        self.assertEqual(
            sorted(expected),
            sorted(self._agent._physical_network_mappings.items())
        )

    @ddt.data(True, False)
    @mock.patch.object(agent_base.Layer2Agent, '_validate_vswitch')
    def test_validate_vswitches(self, all_valid, mock_validate_vswitch):
        phys_mappings = {
            'fakenetwork0': mock.sentinel.vswitch_name_0,
            'fakenetwork1': mock.sentinel.vswitch_name_1
        }
        self._agent._physical_network_mappings = phys_mappings
        self._agent._local_network_vswitch = mock.sentinel.local_vswitch

        exp_vswitch_names = (
            list(phys_mappings.values()) + [mock.sentinel.local_vswitch])

        mock_validate_vswitch.side_effect = (
            exception.ValidationError if not all_valid else None,
            None, None)

        if all_valid:
            self._agent._validate_vswitches()
        else:
            self.assertRaises(exception.ValidationError,
                              self._agent._validate_vswitches)

        mock_validate_vswitch.assert_has_calls(
            [mock.call(vswitch_name)
             for vswitch_name in exp_vswitch_names],
            any_order=True)

    def test_validate_vswitches_none_configured(self):
        self._agent._physical_network_mappings = {}
        self._agent._local_network_vswitch = None

        self.assertRaises(exception.ValidationError,
                          self._agent._validate_vswitches)

    def test_is_ovs_extension(self):
        valid_ovs_ext_names = ['Cloudbase Open vSwitch Extension',
                               'Open vSwitch Extension',
                               'VendorX ovs',
                               'open v-switch',
                               'openvswitch']
        for ovs_ext_name in valid_ovs_ext_names:
            ext = dict(name=ovs_ext_name)
            self.assertTrue(
                self._agent._is_ovs_extension(ext))

        ext = dict(name='fake extension')
        self.assertFalse(self._agent._is_ovs_extension(ext))

    @ddt.data({},
              {'is_valid': False, 'exists': False},
              {'is_valid': False,
               'is_ovs_ext': True,
               'enabled_state': os_win_const.CIM_STATE_ENABLED},
              {'is_ovs_ext': True,
               'enabled_state': os_win_const.CIM_STATE_DISABLED},
              {'enabled_state': os_win_const.CIM_STATE_ENABLED})
    @ddt.unpack
    @mock.patch.object(agent_base.Layer2Agent, '_is_ovs_extension')
    def test_validate_vswitch(self, mock_is_ovs_ext,
                              is_valid=True, exists=True, is_ovs_ext=False,
                              enabled_state=os_win_const.CIM_STATE_ENABLED):
        extension = dict(enabled_state=enabled_state)
        mock_is_ovs_ext.return_value = is_ovs_ext

        mock_get_ext = self._agent._utils.get_vswitch_extensions
        mock_get_ext.side_effect = (
            [[extension]] if exists
            else os_win_exc.HyperVvSwitchNotFound(message='fake_msg'))

        if is_valid:
            self._agent._validate_vswitch(mock.sentinel.vswitch_name)
            mock_is_ovs_ext.assert_called_once_with(extension)
        else:
            self.assertRaises(exception.ValidationError,
                              self._agent._validate_vswitch,
                              mock.sentinel.vswitch_name)

        mock_get_ext.assert_called_once_with(mock.sentinel.vswitch_name)

    def test_get_vswitch_for_physical_network_with_default_switch(self):
        test_mappings = [
            'fakenetwork:fake_vswitch',
            'fakenetwork2$:fake_vswitch_2',
            'fakenetwork*:fake_vswitch_3'
        ]
        self._agent._physical_network_mappings = collections.OrderedDict()
        self._agent._load_physical_network_mappings(test_mappings)
        get_vswitch = self._agent._get_vswitch_for_physical_network

        self.assertEqual('fake_vswitch', get_vswitch('fakenetwork'))
        self.assertEqual('fake_vswitch_2', get_vswitch('fakenetwork2$'))
        self.assertEqual('fake_vswitch_3', get_vswitch('fakenetwork3'))
        self.assertEqual('fake_vswitch_3', get_vswitch('fakenetwork35'))
        self.assertIsNone(get_vswitch('fake_network1'))

    def test_get_vswitch_for_physical_network_without_default_switch(self):
        test_mappings = [
            'fakenetwork:fake_vswitch',
            'fakenetwork2:fake_vswitch_2'
        ]
        self._agent._load_physical_network_mappings(test_mappings)
        get_vswitch = self._agent._get_vswitch_for_physical_network

        self.assertEqual('fake_vswitch', get_vswitch("fakenetwork"))
        self.assertEqual('fake_vswitch_2', get_vswitch("fakenetwork2"))

    def test_get_vswitch_for_physical_network_none(self):
        get_vswitch = self._agent._get_vswitch_for_physical_network

        test_mappings = [
            'fakenetwork:fake_vswitch',
            'fakenetwork2:fake_vswitch_2'
        ]
        self._agent._load_physical_network_mappings(test_mappings)
        self.assertIsNone(get_vswitch(None))

        test_mappings = [
            'fakenetwork:fake_vswitch',
            'fakenetwork2:fake_vswitch_2',
            '*:fake_vswitch_3'
        ]
        self._agent._load_physical_network_mappings(test_mappings)
        self.assertEqual('fake_vswitch_3', get_vswitch(None))

    def test_get_vswitch_name_local(self):
        self._agent._local_network_vswitch = 'test_local_switch'
        ret = self._agent._get_vswitch_name(
            constants.TYPE_LOCAL, mock.sentinel.FAKE_PHYSICAL_NETWORK)

        self.assertEqual('test_local_switch', ret)

    @mock.patch.object(agent_base.Layer2Agent,
                       "_get_vswitch_for_physical_network")
    def test_get_vswitch_name_vlan(self, mock_get_vswitch_for_phys_net):
        ret = self._agent._get_vswitch_name(
            constants.TYPE_VLAN, mock.sentinel.FAKE_PHYSICAL_NETWORK)

        self.assertEqual(mock_get_vswitch_for_phys_net.return_value, ret)
        mock_get_vswitch_for_phys_net.assert_called_once_with(
            mock.sentinel.FAKE_PHYSICAL_NETWORK)

    @mock.patch.object(agent_base.Layer2Agent,
                       "_get_vswitch_for_physical_network")
    @ddt.data(constants.TYPE_VLAN, constants.TYPE_LOCAL)
    def test_get_vswitch_name_missing(self, network_type,
                                      mock_get_vswitch_for_phys_net):
        mock_get_vswitch_for_phys_net.return_value = None
        self._agent._local_network_vswitch = ''

        self.assertRaises(exception.NetworkingHyperVException,
                          self._agent._get_vswitch_name,
                          network_type,
                          mock.sentinel.FAKE_PHYSICAL_NETWORK)

    def test_get_network_vswitch_map_by_port_id(self):
        net_uuid = 'net-uuid'
        self._agent._network_vswitch_map = {
            net_uuid: {'ports': [self._FAKE_PORT_ID]}
        }

        network, port_map = self._agent._get_network_vswitch_map_by_port_id(
            self._FAKE_PORT_ID)

        self.assertEqual(net_uuid, network)
        self.assertEqual({'ports': [self._FAKE_PORT_ID]}, port_map)

    def test_get_network_vswitch_map_by_port_id_not_found(self):
        net_uuid = 'net-uuid'
        self._agent._network_vswitch_map = {net_uuid: {'ports': []}}

        network, port_map = self._agent._get_network_vswitch_map_by_port_id(
            self._FAKE_PORT_ID)

        self.assertIsNone(network)
        self.assertIsNone(port_map)

    def test_update_port_status_cache_added(self):
        self._agent._unbound_ports = set([mock.sentinel.bound_port])
        self._agent._update_port_status_cache(mock.sentinel.bound_port)

        self.assertEqual(set([mock.sentinel.bound_port]),
                         self._agent._bound_ports)
        self.assertEqual(set([]), self._agent._unbound_ports)

    def test_update_port_status_cache_removed(self):
        self._agent._bound_ports = set([mock.sentinel.unbound_port])
        self._agent._update_port_status_cache(mock.sentinel.unbound_port,
                                              device_bound=False)

        self.assertEqual(set([]), self._agent._bound_ports)
        self.assertEqual(set([mock.sentinel.unbound_port]),
                         self._agent._unbound_ports)

    @mock.patch('eventlet.spawn_n')
    def test_create_event_listeners(self, mock_spawn):
        self._agent._event_callback_pairs = [
            (mock.sentinel.event_type, mock.sentinel.callback)]

        self._agent._create_event_listeners()

        self._agent._utils.get_vnic_event_listener.assert_called_once_with(
            mock.sentinel.event_type)
        mock_spawn.assert_called_once_with(
            self._agent._utils.get_vnic_event_listener.return_value,
            mock.sentinel.callback)

    @mock.patch.object(agent_base.Layer2Agent,
                       '_create_event_listeners')
    def test_prologue(self, mock_create_listeners):
        self._agent._prologue()

        # self._added_ports = self._utils.get_vnic_ids()
        mock_create_listeners.assert_called_once_with()

    def test_reclaim_local_network(self):
        self._agent._network_vswitch_map = {}
        self._agent._network_vswitch_map[mock.sentinel.net_id] = (
            mock.sentinel.vswitch)

        self._agent._reclaim_local_network(mock.sentinel.net_id)
        self.assertNotIn(mock.sentinel.net_id,
                         self._agent._network_vswitch_map)

    @mock.patch.object(_Layer2Agent, '_provision_network')
    def test_port_bound_net_type(self, mock_provision_network):
        net_uuid = 'my-net-uuid'
        fake_map = {'vswitch_name': mock.sentinel.vswitch_name,
                    'ports': []}

        def fake_prov_network(*args, **kwargs):
            self._agent._network_vswitch_map[net_uuid] = fake_map

        mock_provision_network.side_effect = fake_prov_network

        self._agent._port_bound(mock.sentinel.port_id,
                                net_uuid, mock.sentinel.network_type,
                                mock.sentinel.physical_network,
                                mock.sentinel.segmentation_id,
                                mock.sentinel.port_security_enabled, True)

        self.assertIn(mock.sentinel.port_id, fake_map['ports'])
        mock_provision_network.assert_called_once_with(
            mock.sentinel.port_id, net_uuid, mock.sentinel.network_type,
            mock.sentinel.physical_network, mock.sentinel.segmentation_id)
        self._agent._utils.connect_vnic_to_vswitch.assert_called_once_with(
            vswitch_name=mock.sentinel.vswitch_name,
            switch_port_name=mock.sentinel.port_id)
        self._agent._utils.set_vswitch_port_sriov.assert_called_once_with(
            mock.sentinel.port_id, True)

    @mock.patch.object(agent_base.Layer2Agent,
                       '_get_network_vswitch_map_by_port_id')
    def _check_port_unbound(self, mock_get_vswitch_map_by_port_id, ports=None,
                            net_uuid=None):
        vswitch_map = {
            'network_type': 'vlan',
            'vswitch_name': 'fake-vswitch',
            'ports': ports,
            'vlan_id': 1}
        network_vswitch_map = (net_uuid, vswitch_map)
        mock_get_vswitch_map_by_port_id.return_value = network_vswitch_map

        with mock.patch.object(
                self._agent._utils,
                'remove_switch_port') as mock_remove_switch_port:
            self._agent._port_unbound(self._FAKE_PORT_ID, vnic_deleted=False)

            if net_uuid:
                mock_remove_switch_port.assert_called_once_with(
                    self._FAKE_PORT_ID, False)
            else:
                self.assertFalse(mock_remove_switch_port.called)

    @mock.patch.object(agent_base.Layer2Agent,
                       '_reclaim_local_network')
    def test_port_unbound(self, mock_reclaim_local_network):
        net_uuid = 'my-net-uuid'
        self._check_port_unbound(ports=[self._FAKE_PORT_ID],
                                 net_uuid=net_uuid)
        mock_reclaim_local_network.assert_called_once_with(net_uuid)

    def test_port_unbound_port_not_found(self):
        self._check_port_unbound()

    @mock.patch.object(_Layer2Agent, '_treat_vif_port')
    def test_process_added_port_sriov(self, mock_treat_vif_port):
        details = self._get_fake_port_details()
        details.pop('device')
        port_details = dict(profile={'pci_slot': mock.sentinel.pci_slot},
                            **self._get_fake_port_details())

        self._agent.process_added_port(port_details)
        mock_treat_vif_port.assert_called_once_with(set_port_sriov=True,
                                                    **details)

    @ddt.data(os_win_exc.HyperVvNicNotFound(vnic_name='fake_vnic'),
              os_win_exc.HyperVPortNotFoundException(port_name='fake_port'),
              Exception)
    @mock.patch.object(_Layer2Agent, '_treat_vif_port')
    def test_process_added_port_failed(self, side_effect, mock_treat_vif_port):
        mock_treat_vif_port.side_effect = side_effect
        self._agent._added_ports = set()
        details = self._get_fake_port_details()
        details.pop('device')
        port_details = self._get_fake_port_details()

        self._agent.process_added_port(port_details)

        if isinstance(side_effect, os_win_exc.HyperVvNicNotFound):
            self.assertNotIn(mock.sentinel.device, self._agent._added_ports)
        else:
            self.assertIn(mock.sentinel.device, self._agent._added_ports)

        mock_treat_vif_port.assert_called_once_with(set_port_sriov=False,
                                                    **details)

    def test_treat_devices_added_returns_true_for_missing_device(self):
        self._agent._added_ports = set([mock.sentinel.port_id])
        attrs = {'get_devices_details_list.side_effect': Exception()}
        self._agent._plugin_rpc.configure_mock(**attrs)
        self._agent._treat_devices_added()

        self.assertIn(mock.sentinel.port_id, self._agent._added_ports)

    @mock.patch('eventlet.spawn_n')
    def test_treat_devices_added_updates_known_port(self, mock_spawn):
        self._agent._added_ports = set([mock.sentinel.device])
        fake_port_details = self._get_fake_port_details()
        kwargs = {'get_devices_details_list.return_value': [fake_port_details]}
        self._agent._plugin_rpc.configure_mock(**kwargs)

        self._agent._treat_devices_added()

        mock_spawn.assert_called_once_with(
            self._agent.process_added_port, fake_port_details)
        self.assertNotIn(mock.sentinel.device, self._agent._added_ports)

    def test_treat_devices_added_missing_port_id(self):
        self._agent._added_ports = set([mock.sentinel.port_id])
        details = {'device': mock.sentinel.port_id}
        attrs = {'get_devices_details_list.return_value': [details]}
        self._agent._plugin_rpc.configure_mock(**attrs)

        self._agent._treat_devices_added()

        self.assertNotIn(mock.sentinel.port_id, self._agent._added_ports)

    @mock.patch.object(agent_base.Layer2Agent,
                       '_port_unbound')
    @mock.patch.object(agent_base.Layer2Agent,
                       '_update_port_status_cache')
    def test_process_removed_port_exception(self, mock_update_port_cache,
                                            mock_port_unbound):
        self._agent._removed_ports = set([mock.sentinel.port_id])

        raised_exc = exception.NetworkingHyperVException
        mock_port_unbound.side_effect = raised_exc

        self.assertRaises(raised_exc,
                          self._agent._process_removed_port,
                          mock.sentinel.port_id)

        mock_update_port_cache.assert_called_once_with(
            mock.sentinel.port_id, device_bound=False)
        self.assertIn(mock.sentinel.port_id, self._agent._removed_ports)

    @mock.patch.object(agent_base.Layer2Agent,
                       '_port_unbound')
    @mock.patch.object(agent_base.Layer2Agent,
                       '_update_port_status_cache')
    def test_process_removed_port(self, mock_update_port_cache,
                                  mock_port_unbound):
        self._agent._removed_ports = set([mock.sentinel.port_id])

        self._agent._process_removed_port(mock.sentinel.port_id)

        mock_update_port_cache.assert_called_once_with(
            mock.sentinel.port_id, device_bound=False)
        mock_port_unbound.assert_called_once_with(mock.sentinel.port_id,
                                                  vnic_deleted=True)
        self.assertNotIn(mock.sentinel.port_id, self._agent._removed_ports)

    @mock.patch('eventlet.spawn_n')
    def test_treat_devices_removed(self, mock_spawn):
        mock_removed_ports = [mock.sentinel.port0, mock.sentinel.port1]
        self._agent._removed_ports = set(mock_removed_ports)

        self._agent._treat_devices_removed()

        mock_spawn.assert_has_calls(
            [mock.call(self._agent._process_removed_port, port)
             for port in mock_removed_ports],
            any_order=True)

    def test_notify_plugin_no_updates(self):
        self._agent._bound_ports = set()
        self._agent._unbound_ports = set()

        self._agent._notify_plugin_on_port_updates()

        self.assertFalse(self._agent._plugin_rpc.update_device_list.called)

    def test_notify_plugin(self):
        self._agent._bound_ports = set([mock.sentinel.bound_port])
        self._agent._unbound_ports = set([mock.sentinel.unbound_port])

        self._agent._notify_plugin_on_port_updates()

        self._agent._plugin_rpc.update_device_list.assert_called_once_with(
            self._agent._context, [mock.sentinel.bound_port],
            [mock.sentinel.unbound_port], self._agent._agent_id,
            self._agent._host)
        self.assertEqual(set([]), self._agent._bound_ports)
        self.assertEqual(set([]), self._agent._unbound_ports)

    @mock.patch.object(agent_base.Layer2Agent, '_treat_devices_removed')
    @mock.patch.object(agent_base.Layer2Agent, '_treat_devices_added')
    @mock.patch('eventlet.spawn_n')
    def test_work(self, mock_spawn, mock_treat_dev_added,
                  mock_treat_dev_removed):
        self._agent._refresh_cache = True
        self._agent._bound_ports = set([mock.sentinel.bound_port])
        self._agent._added_ports = set([mock.sentinel.bound_port])
        self._agent._removed_ports = set([mock.sentinel.unbound_port])

        self._agent._work()

        self._agent._utils.update_cache.assert_called_once_with()
        self.assertFalse(self._agent._refresh_cache)
        mock_spawn.assert_called_once_with(
            self._agent._notify_plugin_on_port_updates)
        mock_treat_dev_added.assert_called_once_with()
        mock_treat_dev_removed.assert_called_once_with()

    @mock.patch.object(agent_base.Layer2Agent, '_treat_devices_removed')
    @mock.patch.object(agent_base.Layer2Agent, '_treat_devices_added')
    @mock.patch('eventlet.spawn_n')
    def test_work_noop(self, mock_spawn, mock_treat_dev_added,
                       mock_treat_dev_removed):
        self._agent._work()

        self.assertFalse(mock_spawn.called)
        self.assertFalse(mock_treat_dev_added.called)
        self.assertFalse(mock_treat_dev_removed.called)

    def test_port_update_not_found(self):
        self._agent._utils.vnic_port_exists.return_value = False
        port = {'id': mock.sentinel.port_id}
        self._agent.port_update(self._agent._context, port)

    @mock.patch.object(agent_base.Layer2Agent, '_treat_vif_port')
    def test_port_update(self, mock_treat_vif_port):
        self._agent._utils.vnic_port_exists.return_value = True
        port = {'id': mock.sentinel.port_id,
                'network_id': mock.sentinel.network_id,
                'admin_state_up': mock.sentinel.admin_state_up,
                'port_security_enabled': mock.sentinel.port_security_enabled}

        self._agent.port_update(self._agent._context, port,
                                mock.sentinel.network_type,
                                mock.sentinel.segmentation_id,
                                mock.sentinel.physical_network)

    @mock.patch.object(agent_base.Layer2Agent,
                       '_reclaim_local_network')
    def test_network_delete(self, mock_reclaim_local_network):
        self._agent._network_vswitch_map = {}
        self._agent._network_vswitch_map[mock.sentinel.net_id] = (
            mock.sentinel.vswitch)

        self._agent.network_delete(mock.sentinel.context, mock.sentinel.net_id)
        mock_reclaim_local_network.assert_called_once_with(
            mock.sentinel.net_id)

    @mock.patch.object(agent_base.Layer2Agent,
                       '_reclaim_local_network')
    def test_network_delete_not_defined(self, mock_reclaim_local_network):
        self._agent.network_delete(mock.sentinel.context, mock.sentinel.net_id)
        self.assertFalse(mock_reclaim_local_network.called)

    @mock.patch.object(agent_base.Layer2Agent, '_port_bound')
    def test_treat_vif_port_state_up(self, mock_port_bound):
        self._agent._treat_vif_port(
            mock.sentinel.port_id, mock.sentinel.network_id,
            mock.sentinel.network_type, mock.sentinel.physical_network,
            mock.sentinel.segmentation_id, True,
            mock.sentinel.port_security_enabled)

        mock_port_bound.assert_called_once_with(
            mock.sentinel.port_id, mock.sentinel.network_id,
            mock.sentinel.network_type, mock.sentinel.physical_network,
            mock.sentinel.segmentation_id, mock.sentinel.port_security_enabled,
            False)

    @mock.patch.object(agent_base.Layer2Agent, '_port_unbound')
    def test_treat_vif_port_state_down(self, mock_port_unbound):
        self._agent._treat_vif_port(
            mock.sentinel.port_id, mock.sentinel.network_id,
            mock.sentinel.network_type, mock.sentinel.physical_network,
            mock.sentinel.segmentation_id, False,
            mock.sentinel.port_security_enabled)

        mock_port_unbound.assert_called_once_with(mock.sentinel.port_id)



# Copyright 2015 Cloudbase Solutions SRL
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Unit tests for the neutron client.
"""

import mock

from networking_hyperv.neutron import config
from networking_hyperv.neutron import constants
from networking_hyperv.neutron import neutron_client
from networking_hyperv.tests import base

CONF = config.CONF


class TestNeutronClient(base.BaseTestCase):

    _autospec_classes = [
        neutron_client.clientv20.Client,
    ]

    _FAKE_CIDR = '10.0.0.0/24'
    _FAKE_GATEWAY = '10.0.0.1'
    _FAKE_HOST = 'fake_host'

    def setUp(self):
        super(TestNeutronClient, self).setUp()
        self._neutron = neutron_client.NeutronAPIClient()

    @mock.patch.object(neutron_client.clientv20, "Client")
    @mock.patch.object(neutron_client, "ks_loading")
    def test_init_client(self, mock_ks_loading, mock_client):
        self._neutron._init_client()

        self.assertEqual(mock_client.return_value, self._neutron._client)
        mock_ks_loading.load_session_from_conf_options.assert_called_once_with(
            CONF, config.NEUTRON_GROUP)
        mock_ks_loading.load_auth_from_conf_options.assert_called_once_with(
            CONF, config.NEUTRON_GROUP)
        session = mock_ks_loading.load_session_from_conf_options.return_value
        plugin = mock_ks_loading.load_auth_from_conf_options.return_value
        mock_client.assert_called_once_with(
            session=session,
            auth=plugin)

    def test_get_network_subnets(self):
        self._neutron._client.show_network.return_value = {
            'network': {
                'subnets': [mock.sentinel.fake_subnet]
            }
        }

        subnets = self._neutron.get_network_subnets(mock.sentinel.net_id)

        self._neutron._client.show_network.assert_called_once_with(
            mock.sentinel.net_id)
        self.assertEqual([mock.sentinel.fake_subnet], subnets)

    def test_get_network_subnets_exception(self):
        self._neutron._client.show_network.side_effect = Exception("Fail")
        subnets = self._neutron.get_network_subnets(mock.sentinel.net_id)
        self.assertEqual([], subnets)

    def test_get_network_subnet_cidr(self):
        self._neutron._client.show_subnet.return_value = {
            'subnet': {
                'cidr': self._FAKE_CIDR,
                'gateway_ip': self._FAKE_GATEWAY,
            }
        }

        cidr, gw = self._neutron.get_network_subnet_cidr_and_gateway(
            mock.sentinel.subnet_id)

        self._neutron._client.show_subnet.assert_called_once_with(
            mock.sentinel.subnet_id)
        self.assertEqual(self._FAKE_CIDR, cidr)
        self.assertEqual(self._FAKE_GATEWAY, gw)

    def test_get_network_subnet_cidr_exception(self):
        self._neutron._client.show_subnet.side_effect = Exception("Fail")
        cidr, gw = self._neutron.get_network_subnet_cidr_and_gateway(
            mock.sentinel.subnet_id)
        self.assertIsNone(cidr)
        self.assertIsNone(gw)

    def test_get_port_ip_address(self):
        self._neutron._client.show_port.return_value = {
            'port': {
                'fixed_ips': [{'ip_address': mock.sentinel.ip_addr}]
            }
        }

        ip_addr = self._neutron.get_port_ip_address(mock.sentinel.fake_port_id)

        self._neutron._client.show_port.assert_called_once_with(
            mock.sentinel.fake_port_id)
        self.assertEqual(mock.sentinel.ip_addr, ip_addr)

    def test_get_port_ip_address_exception(self):
        self._neutron._client.show_port.side_effect = Exception("Fail")
        ip_addr = self._neutron.get_port_ip_address(mock.sentinel.fake_port_id)
        self.assertIsNone(ip_addr)

    def test_get_tunneling_agents(self):
        non_tunnel_agent = {}
        ignored_agent = {'configurations': {
            'tunnel_types': [constants.TYPE_NVGRE]}
        }
        tunneling_agent = {
            'configurations': {'tunnel_types': [constants.TYPE_NVGRE],
                               'tunneling_ip': mock.sentinel.tunneling_ip},
            'host': self._FAKE_HOST
        }

        self._neutron._client.list_agents.return_value = {
            'agents': [non_tunnel_agent, ignored_agent, tunneling_agent]
        }

        actual = self._neutron.get_tunneling_agents()
        self.assertEqual({self._FAKE_HOST: mock.sentinel.tunneling_ip}, actual)

    def test_get_tunneling_agents_exception(self):
        self._neutron._client.list_agents.side_effect = Exception("Fail")
        actual = self._neutron.get_tunneling_agents()
        self.assertEqual({}, actual)

    def test_get_network_ports(self):
        self._neutron._client.list_ports.return_value = {
            'ports': [mock.sentinel.port]
        }

        actual = self._neutron.get_network_ports(key='value')

        self._neutron._client.list_ports.assert_called_once_with(key='value')
        self.assertEqual([mock.sentinel.port], actual)

    def test_get_network_ports_exception(self):
        self._neutron._client.list_ports.side_effect = Exception("Fail")
        actual = self._neutron.get_network_ports()
        self.assertEqual([], actual)

    def test_get_port_profile_id(self):
        fake_profile_id = 'fake_profile_id'
        self._neutron._client.show_port.return_value = {
            'port': {
                'binding:vif_details': {'port_profile_id': fake_profile_id}
            }
        }

        actual = self._neutron.get_port_profile_id(mock.sentinel.port_id)

        self.assertEqual('{%s}' % fake_profile_id, actual)
        self._neutron._client.show_port.assert_called_once_with(
            mock.sentinel.port_id)

    def test_get_port_profile_id_failed(self):
        self._neutron._client.show_port.side_effect = Exception("Fail")
        actual = self._neutron.get_port_profile_id(mock.sentinel.port_id)

        self.assertEqual({}, actual)
        self._neutron._client.show_port.assert_called_once_with(
            mock.sentinel.port_id)



# Copyright 2015 Cloudbase Solutions SRL
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Unit Tests for Hyper-V Agent Notifier.
"""

import mock

from networking_hyperv.neutron import constants
from networking_hyperv.neutron import hyperv_agent_notifier
from networking_hyperv.tests import base


class TestAgentNotifierApi(base.BaseTestCase):

    def setUp(self):
        super(TestAgentNotifierApi, self).setUp()

        self.notifier = hyperv_agent_notifier.AgentNotifierApi(
            topic=constants.AGENT_TOPIC, client=mock.MagicMock())

    def test_tunnel_update(self):
        expected_topic = hyperv_agent_notifier.get_topic_name(
            constants.AGENT_TOPIC, constants.TUNNEL, constants.UPDATE)

        self.notifier.tunnel_update(mock.sentinel.context,
                                    mock.sentinel.tunnel_ip,
                                    constants.TYPE_NVGRE)

        self.notifier._client.prepare.assert_called_once_with(
            topic=expected_topic, fanout=True)
        prepared_client = self.notifier._client.prepare.return_value
        prepared_client.cast.assert_called_once_with(
            mock.sentinel.context, 'tunnel_update',
            tunnel_ip=mock.sentinel.tunnel_ip,
            tunnel_type=constants.TYPE_NVGRE)

    def test_lookup_update(self):
        expected_topic = hyperv_agent_notifier.get_topic_name(
            constants.AGENT_TOPIC, constants.LOOKUP, constants.UPDATE)

        self.notifier.lookup_update(mock.sentinel.context,
                                    mock.sentinel.lookup_ip,
                                    mock.sentinel.lookup_details)

        self.notifier._client.prepare.assert_called_once_with(
            topic=expected_topic, fanout=True)
        prepared_client = self.notifier._client.prepare.return_value
        prepared_client.cast.assert_called_once_with(
            mock.sentinel.context, 'lookup_update',
            lookup_ip=mock.sentinel.lookup_ip,
            lookup_details=mock.sentinel.lookup_details)



# Copyright 2017 Cloudbase Solutions Srl
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Unit tests for the Hyper-V Trunk Driver.
"""

import mock
from neutron.api.rpc.callbacks import events
from neutron.api.rpc.handlers import resources_rpc
from neutron_lib.services.trunk import constants as t_const
from os_win import constants as os_win_const
import oslo_messaging
import testtools

from networking_hyperv.neutron import trunk_driver
from networking_hyperv.tests import base


class TestHyperVTrunkDriver(base.HyperVBaseTestCase):

    _autospec_classes = [
        trunk_driver.trunk_rpc.TrunkStub,
    ]

    @mock.patch.object(trunk_driver.trunk_rpc.TrunkSkeleton, '__init__',
                       lambda *args, **kwargs: None)
    def setUp(self):
        super(TestHyperVTrunkDriver, self).setUp()

        self.trunk_driver = trunk_driver.HyperVTrunkDriver(
            mock.sentinel.context)
        self.trunk_driver._utils = mock.MagicMock(
            autospec=self.trunk_driver._utils)

    def test_handle_trunks_deleted(self):
        mock_trunk = mock.MagicMock()
        self.trunk_driver._trunks[mock_trunk.id] = mock_trunk

        self.trunk_driver.handle_trunks([mock_trunk], events.DELETED)
        self.assertNotIn(mock_trunk.id, self.trunk_driver._trunks)

    @mock.patch.object(trunk_driver.HyperVTrunkDriver, '_setup_trunk')
    def test_handle_trunks_created(self, mock_setup_trunk):
        sub_ports = []
        mock_trunk = mock.MagicMock(sub_ports=sub_ports)

        self.trunk_driver.handle_trunks([mock_trunk], events.CREATED)

        self.assertEqual(mock_trunk, self.trunk_driver._trunks[mock_trunk.id])
        mock_setup_trunk.assert_called_once_with(mock_trunk)

    @mock.patch.object(trunk_driver.HyperVTrunkDriver, '_set_port_vlan')
    @mock.patch.object(trunk_driver.HyperVTrunkDriver, '_fetch_trunk')
    def test_bind_vlan_port_not_trunk(self, mock_fetch_trunk, mock_set_vlan):
        mock_fetch_trunk.return_value = None

        self.trunk_driver.bind_vlan_port(mock.sentinel.port_id,
                                         mock.sentinel.segmentation_id)

        mock_fetch_trunk.assert_called_once_with(mock.sentinel.port_id)
        mock_set_vlan.assert_called_once_with(mock.sentinel.port_id,
                                              mock.sentinel.segmentation_id)

    @mock.patch.object(trunk_driver.HyperVTrunkDriver, '_setup_trunk')
    @mock.patch.object(trunk_driver.HyperVTrunkDriver, '_fetch_trunk')
    def test_bind_vlan_port(self, mock_fetch_trunk, mock_setup_trunk):
        self.trunk_driver.bind_vlan_port(mock.sentinel.port_id,
                                         mock.sentinel.segmentation_id)

        mock_fetch_trunk.assert_called_once_with(mock.sentinel.port_id)
        mock_setup_trunk.assert_called_once_with(mock_fetch_trunk.return_value,
                                                 mock.sentinel.segmentation_id)

    def test_fetch_trunk(self):
        mock_trunk = (
            self.trunk_driver._trunk_rpc.get_trunk_details.return_value)

        trunk = self.trunk_driver._fetch_trunk(mock.sentinel.port_id,
                                               mock.sentinel.context)

        self.assertEqual(mock_trunk, trunk)
        self.assertEqual(mock_trunk, self.trunk_driver._trunks[mock_trunk.id])
        self.trunk_driver._trunk_rpc.get_trunk_details.assert_called_once_with(
            mock.sentinel.context, mock.sentinel.port_id)

    def test_fetch_trunk_resource_not_found(self):
        self.trunk_driver._trunk_rpc.get_trunk_details.side_effect = (
            resources_rpc.ResourceNotFound)

        trunk = self.trunk_driver._fetch_trunk(mock.sentinel.port_id)
        self.assertIsNone(trunk)

    def test_fetch_trunk_resource_remote_error(self):
        self.trunk_driver._trunk_rpc.get_trunk_details.side_effect = (
            oslo_messaging.RemoteError('expected CallbackNotFound'))

        trunk = self.trunk_driver._fetch_trunk(mock.sentinel.port_id)
        self.assertIsNone(trunk)

    def test_fetch_trunk_resource_remote_error_reraised(self):
        self.trunk_driver._trunk_rpc.get_trunk_details.side_effect = (
            oslo_messaging.RemoteError)

        self.assertRaises(oslo_messaging.RemoteError,
                          self.trunk_driver._fetch_trunk,
                          mock.sentinel.port_id)

    @mock.patch.object(trunk_driver.HyperVTrunkDriver, '_set_port_vlan')
    def test_setup_trunk(self, mock_set_vlan):
        mock_subport = mock.MagicMock()
        mock_trunk = mock.MagicMock(sub_ports=[mock_subport])
        trunk_rpc = self.trunk_driver._trunk_rpc
        trunk_rpc.update_trunk_status.side_effect = [
            testtools.ExpectedException, None]

        self.trunk_driver._setup_trunk(mock_trunk, mock.sentinel.vlan_id)

        trunk_rpc.update_subport_bindings.assert_called_once_with(
            self.trunk_driver._context, [mock_subport])
        mock_set_vlan.assert_called_once_with(
            mock_trunk.port_id, mock.sentinel.vlan_id,
            [mock_subport.segmentation_id])
        mock_set_vlan.has_calls([
            mock.call(self.trunk_driver._context, mock_trunk.id, status)
            for status in [t_const.TRUNK_ACTIVE_STATUS,
                           t_const.TRUNK_DEGRADED_STATUS]])

    def _check_set_port_vlan(self, vlan_trunk, operation_mode):
        self.trunk_driver._set_port_vlan(mock.sentinel.port_id,
                                         mock.sentinel.vlan_id,
                                         vlan_trunk)

        self.trunk_driver._utils.set_vswitch_port_vlan_id(
            mock.sentinel.vlan_id, mock.sentinel.port_id,
            operation_mode=operation_mode,
            trunk_vlans=vlan_trunk)

    def test_set_port_vlan_trunk_mode(self):
        self._check_set_port_vlan(mock.sentinel.vlan_trunk,
                                  os_win_const.VLAN_MODE_TRUNK)

    def test_set_port_vlan_access_mode(self):
        self._check_set_port_vlan(None, os_win_const.VLAN_MODE_ACCESS)



# Copyright 2014 Cloudbase Solutions SRL
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import threading

import netaddr
from neutron.agent import firewall
from neutron_lib import constants
from os_win import exceptions
from os_win.utils.network import networkutils
from os_win import utilsfactory
from oslo_log import log as logging
import six

from networking_hyperv.common.i18n import _LE, _LI  # noqa
from networking_hyperv.neutron import _common_utils as c_utils

LOG = logging.getLogger(__name__)

DIRECTION_IP_PREFIX = {constants.INGRESS_DIRECTION: 'source_ip_prefix',
                       constants.EGRESS_DIRECTION: 'dest_ip_prefix'}

ACL_PROP_MAP = {
    'direction': {'ingress': networkutils.NetworkUtils._ACL_DIR_IN,
                  'egress': networkutils.NetworkUtils._ACL_DIR_OUT},
    'ethertype': {'IPv4': networkutils.NetworkUtils._ACL_TYPE_IPV4,
                  'IPv6': networkutils.NetworkUtils._ACL_TYPE_IPV6},
    'protocol': {'tcp': networkutils.NetworkUtils._TCP_PROTOCOL,
                 'udp': networkutils.NetworkUtils._UDP_PROTOCOL,
                 'icmp': networkutils.NetworkUtils._ICMP_PROTOCOL,
                 'ipv6-icmp': networkutils.NetworkUtils._ICMPV6_PROTOCOL,
                 'icmpv6': networkutils.NetworkUtils._ICMPV6_PROTOCOL},
    'action': {'allow': networkutils.NetworkUtils._ACL_ACTION_ALLOW,
               'deny': networkutils.NetworkUtils._ACL_ACTION_DENY},
    'default': "ANY",
    'address_default': {'IPv4': '0.0.0.0/0', 'IPv6': '::/0'}
}


_ports_synchronized = c_utils.get_port_synchronized_decorator('n-hv-driver-')


class HyperVSecurityGroupsDriverMixin(object):
    """Security Groups Driver.

    Security Groups implementation for Hyper-V VMs.
    """

    def __init__(self):
        self._utils = utilsfactory.get_networkutils()
        self._sg_gen = SecurityGroupRuleGeneratorR2()
        self._sec_group_rules = {}
        self._security_ports = {}
        self._sg_members = {}
        self._sg_rule_templates = {}
        self.cache_lock = threading.Lock()

        # TODO(claudiub): remove this on the next os-win release.
        clear_cache = lambda port_id: self._utils._sg_acl_sds.pop(port_id,
                                                                  None)
        self._utils.clear_port_sg_acls_cache = clear_cache

    def _select_sg_rules_for_port(self, port, direction):
        sg_ids = port.get('security_groups', [])
        port_rules = []
        fixed_ips = port.get('fixed_ips', [])
        for sg_id in sg_ids:
            for rule in self._sg_rule_templates.get(sg_id, []):
                if rule['direction'] != direction:
                    continue
                remote_group_id = rule.get('remote_group_id')
                if not remote_group_id:
                    grp_rule = rule.copy()
                    grp_rule.pop('security_group_id', None)
                    port_rules.append(grp_rule)
                    continue
                ethertype = rule['ethertype']
                for ip in self._sg_members[remote_group_id][ethertype]:
                    if ip in fixed_ips:
                        continue
                    ip_rule = rule.copy()
                    direction_ip_prefix = DIRECTION_IP_PREFIX[direction]
                    ip_rule[direction_ip_prefix] = str(
                        netaddr.IPNetwork(ip).cidr)
                    # NOTE(claudiub): avoid returning fields that are not
                    # directly used in setting the security group rules
                    # properly (remote_group_id, security_group_id), as they
                    # only make testing for rule's identity harder.
                    ip_rule.pop('security_group_id', None)
                    ip_rule.pop('remote_group_id', None)
                    port_rules.append(ip_rule)
        return port_rules

    def filter_defer_apply_on(self):
        """Defer application of filtering rule."""
        pass

    def filter_defer_apply_off(self):
        """Turn off deferral of rules and apply the rules now."""
        pass

    def update_security_group_rules(self, sg_id, sg_rules):
        LOG.debug("Update rules of security group (%s)", sg_id)
        with self.cache_lock:
            self._sg_rule_templates[sg_id] = sg_rules

    def update_security_group_members(self, sg_id, sg_members):
        LOG.debug("Update members of security group (%s)", sg_id)
        with self.cache_lock:
            self._sg_members[sg_id] = sg_members

    def _generate_rules(self, ports):
        newports = {}
        for port in ports:
            _rules = []
            _rules.extend(self._select_sg_rules_for_port(
                port, constants.INGRESS_DIRECTION))
            _rules.extend(self._select_sg_rules_for_port(
                port, constants.EGRESS_DIRECTION))
            newports[port['id']] = _rules
        return newports

    @c_utils.ignore_missing_ports
    def prepare_port_filter(self, port):
        if not port.get('port_security_enabled'):
            LOG.info('Port %s does not have security enabled. '
                     'Skipping rules creation.', port['id'])
            return
        LOG.debug('Creating port %s rules', len(port['security_group_rules']))

        # newly created port, add default rules.
        if port['device'] not in self._security_ports:
            LOG.debug('Creating default reject rules.')
            self._sec_group_rules[port['id']] = []

            def_sg_rules = self._sg_gen.create_default_sg_rules()
            self._add_sg_port_rules(port, def_sg_rules)
            # Add provider rules
            provider_rules = port['security_group_rules']
            self._create_port_rules(port, provider_rules)

        newrules = self._generate_rules([port])
        self._create_port_rules(port, newrules[port['id']])

        self._security_ports[port['device']] = port
        self._sec_group_rules[port['id']] = newrules[port['id']]

    @_ports_synchronized
    def _create_port_rules(self, port, rules):
        sg_rules = self._sg_gen.create_security_group_rules(rules)
        old_sg_rules = self._sec_group_rules[port['id']]
        add, rm = self._sg_gen.compute_new_rules_add(old_sg_rules, sg_rules)

        self._add_sg_port_rules(port, list(set(add)))
        self._remove_sg_port_rules(port, list(set(rm)))

    @_ports_synchronized
    def _remove_port_rules(self, port, rules):
        sg_rules = self._sg_gen.create_security_group_rules(rules)
        self._remove_sg_port_rules(port, list(set(sg_rules)))

    def _add_sg_port_rules(self, port, sg_rules):
        if not sg_rules:
            return
        old_sg_rules = self._sec_group_rules[port['id']]
        try:
            self._utils.create_security_rules(port['id'], sg_rules)
            old_sg_rules.extend(sg_rules)
        except exceptions.NotFound:
            # port no longer exists.
            # NOTE(claudiub): In the case of a rebuild / shelve, the
            # neutron port is not deleted, and it can still be in the cache.
            # We need to make sure the port's caches are cleared since it is
            # not valid anymore. The port will be reprocessed in the next
            # loop iteration.
            self._sec_group_rules.pop(port['id'], None)
            self._security_ports.pop(port.get('device'), None)
            raise
        except Exception:
            LOG.exception('Exception encountered while adding rules for '
                          'port: %s', port['id'])
            raise

    def _remove_sg_port_rules(self, port, sg_rules):
        if not sg_rules:
            return
        old_sg_rules = self._sec_group_rules[port['id']]
        try:
            self._utils.remove_security_rules(port['id'], sg_rules)
            for rule in sg_rules:
                if rule in old_sg_rules:
                    old_sg_rules.remove(rule)
        except exceptions.NotFound:
            # port no longer exists.
            self._sec_group_rules.pop(port['id'], None)
            self._security_ports.pop(port.get('device'), None)
            raise
        except Exception:
            LOG.exception('Exception encountered while removing rules for '
                          'port: %s', port['id'])
            raise

    def apply_port_filter(self, port):
        LOG.info('Applying port filter.')

    @c_utils.ignore_missing_ports
    def update_port_filter(self, port):
        if not port.get('port_security_enabled'):
            LOG.info('Port %s does not have security enabled. '
                     'Removing existing rules if any.', port['id'])
            self._security_ports.pop(port.get('device'), None)
            self._sec_group_rules.pop(port['id'], None)
            self._utils.remove_all_security_rules(port['id'])
            return
        LOG.info('Updating port rules.')

        if port['device'] not in self._security_ports:
            LOG.info("Device %(port)s not yet added. Adding.",
                     {'port': port['id']})
            self.prepare_port_filter(port)
            return

        old_port = self._security_ports[port['device']]
        old_provider_rules = old_port['security_group_rules']
        added_provider_rules = port['security_group_rules']
        # Generate the rules
        added_rules = self._generate_rules([port])
        # Expand wildcard rules
        expanded_rules = self._sg_gen.expand_wildcard_rules(
            added_rules[port['id']])
        # Consider added provider rules (if any)
        new_rules = [r for r in added_provider_rules
                     if r not in old_provider_rules]
        # Build new rules to add
        new_rules.extend([r for r in added_rules[port['id']]
                          if r not in self._sec_group_rules[port['id']]])
        # Remove non provider rules
        remove_rules = [r for r in self._sec_group_rules[port['id']]
                        if r not in added_rules[port['id']]]
        # Remove for non provider rules
        remove_rules.extend([r for r in old_provider_rules
                             if r not in added_provider_rules])
        # Avoid removing or adding rules which are contained in wildcard rules
        new_rules = [r for r in new_rules if r not in expanded_rules]
        remove_rules = [r for r in remove_rules if r not in expanded_rules]

        LOG.info("Creating %(new)s new rules, removing %(old)s old rules.",
                 {'new': len(new_rules),
                  'old': len(remove_rules)})

        self._create_port_rules(port, new_rules)
        self._remove_port_rules(old_port, remove_rules)

        self._security_ports[port['device']] = port
        self._sec_group_rules[port['id']] = added_rules[port['id']]

    @c_utils.ignore_missing_ports
    def remove_port_filter(self, port):
        LOG.info('Removing port filter')
        self._security_ports.pop(port['device'], None)
        self._sec_group_rules.pop(port['id'], None)
        self._utils.clear_port_sg_acls_cache(port['id'])

    def security_group_updated(self, action_type, sec_group_ids,
                               device_id=None):
        pass

    @property
    def ports(self):
        return self._security_ports


class SecurityGroupRuleGenerator(object):

    def create_security_group_rules(self, rules):
        security_group_rules = []
        for rule in rules:
            security_group_rules.extend(self.create_security_group_rule(rule))
        return security_group_rules

    def create_security_group_rule(self, rule):
        # TODO(claudiub): implement
        pass

    def _get_rule_remote_address(self, rule):
        if rule['direction'] == 'ingress':
            ip_prefix = 'source_ip_prefix'
        else:
            ip_prefix = 'dest_ip_prefix'

        if ip_prefix in rule:
            return rule[ip_prefix]
        return ACL_PROP_MAP['address_default'][rule['ethertype']]


class SecurityGroupRuleGeneratorR2(SecurityGroupRuleGenerator):

    def create_security_group_rule(self, rule):
        local_port = self._get_rule_port_range(rule)
        direction = ACL_PROP_MAP['direction'][rule['direction']]
        remote_address = self._get_rule_remote_address(rule)
        remote_address = remote_address.split('/128', 1)[0]
        protocol = self._get_rule_protocol(rule)
        if protocol == ACL_PROP_MAP['default']:
            # ANY protocols must be split up, to make stateful rules.
            protocols = list(set(ACL_PROP_MAP['protocol'].values()))
        else:
            protocols = [protocol]

        sg_rules = [SecurityGroupRuleR2(direction=direction,
                                        local_port=local_port,
                                        protocol=proto,
                                        remote_addr=remote_address)
                    for proto in protocols]

        return sg_rules

    def create_default_sg_rules(self):
        ip_type_pairs = [(ACL_PROP_MAP['ethertype'][ip],
                          ACL_PROP_MAP['address_default'][ip])
                         for ip in six.iterkeys(ACL_PROP_MAP['ethertype'])]

        action = ACL_PROP_MAP['action']['deny']
        port = ACL_PROP_MAP['default']
        sg_rules = []
        for direction in ACL_PROP_MAP['direction'].values():
            for protocol in set(ACL_PROP_MAP['protocol'].values()):
                for acl_type, address in ip_type_pairs:
                    sg_rules.append(SecurityGroupRuleR2(direction=direction,
                                                        local_port=port,
                                                        protocol=protocol,
                                                        remote_addr=address,
                                                        action=action))
        return sg_rules

    def compute_new_rules_add(self, old_rules, new_rules):
        add_rules = [r for r in new_rules if r not in old_rules]
        return add_rules, []

    def expand_wildcard_rules(self, rules):
        wildcard_rules = [
            r for r in rules
            if self._get_rule_protocol(r) == ACL_PROP_MAP['default']]
        rules = []
        for r in wildcard_rules:
            rule_copy = r.copy()
            if rule_copy['direction'] == 'ingress':
                ip_prefix = 'source_ip_prefix'
            else:
                ip_prefix = 'dest_ip_prefix'
            if ip_prefix not in rule_copy:
                rule_copy[ip_prefix] = (
                    ACL_PROP_MAP['address_default'][rule_copy['ethertype']])
            for proto in list(set(ACL_PROP_MAP['protocol'].keys())):
                rule_to_add = rule_copy.copy()
                rule_to_add['protocol'] = proto
                rules.extend([rule_to_add])
        return rules

    def _get_rule_port_range(self, rule):
        if 'port_range_min' in rule and 'port_range_max' in rule:
            return '%s-%s' % (rule['port_range_min'],
                              rule['port_range_max'])
        return ACL_PROP_MAP['default']

    def _get_rule_protocol(self, rule):
        protocol = self._get_rule_prop_or_default(rule, 'protocol')
        if protocol == 'icmp' and rule.get('ethertype') == 'IPv6':
            # If protocol is ICMP and ethertype is IPv6 the protocol has
            # to be ICMPv6.
            return ACL_PROP_MAP['protocol']['ipv6-icmp']
        if protocol in six.iterkeys(ACL_PROP_MAP['protocol']):
            return ACL_PROP_MAP['protocol'][protocol]

        return protocol

    def _get_rule_prop_or_default(self, rule, prop):
        if prop in rule:
            return rule[prop]
        return ACL_PROP_MAP['default']


class SecurityGroupRuleBase(object):

    _FIELDS = []

    def __eq__(self, obj):
        for f in self._FIELDS:
            if not hasattr(obj, f) or getattr(obj, f) != getattr(self, f):
                return False
        return True

    def __str__(self):
        return str(self.to_dict())

    def __repr__(self):
        return str(self)

    def to_dict(self):
        return dict((field, getattr(self, field)) for field in self._FIELDS)


class SecurityGroupRuleR2(SecurityGroupRuleBase):

    _FIELDS = ["Direction", "Action", "LocalPort", "Protocol",
               "RemoteIPAddress", "Stateful", "IdleSessionTimeout"]

    IdleSessionTimeout = 0
    Weight = 65500

    def __init__(self, direction, local_port, protocol, remote_addr,
                 action=ACL_PROP_MAP['action']['allow']):
        is_not_icmp = protocol not in [ACL_PROP_MAP['protocol']['icmp'],
                                       ACL_PROP_MAP['protocol']['ipv6-icmp']]

        self.Direction = direction
        self.Action = action
        self.LocalPort = str(local_port) if is_not_icmp else ''
        self.Protocol = protocol
        self.RemoteIPAddress = remote_addr
        self.Stateful = (is_not_icmp and
                         action is not ACL_PROP_MAP['action']['deny'])

        self._cached_hash = hash((direction, action, self.LocalPort,
                                  protocol, remote_addr))

    def __lt__(self, obj):
        return self.Protocol > obj.Protocol

    def __hash__(self):
        return self._cached_hash


class HyperVSecurityGroupsDriver(HyperVSecurityGroupsDriverMixin,
                                 firewall.FirewallDriver):
    pass
